import ast
import os


class StrategyKwargsKeys:
    def __init__(self):
        self.kwargs_list = list()
        self._populate_kwargs_list()

    def _populate_kwargs_list(self):
        _KWARGS_KEYS = list()
        if os.getenv('KWARGS', None):
            _KWARGS = ast.literal_eval(os.getenv('KWARGS', None))
            if _KWARGS is not None:
                for _kwarg in _KWARGS:
                    print(_kwarg)
                    _KWARGS_KEYS.append(_kwarg)
        self.kwargs_list = _KWARGS_KEYS

    def get_kwargs_list(self):
        return self.kwargs_list


strategy_kwargs_list = StrategyKwargsKeys()



import json
import logging
from datetime import date
from aq_lib.event.event import StrategyStatusChangeEvent
from aq_lib.event.event_type import EventType
from aq_lib.model.strategy_status import StrategyStatus
from backtrader import Strategy as btStrategy, LineIterator
from arquants.definitions import sheet_queue, aqhandler
from arquants import sheet_envs
from arquants.event.event_factory import eventFactory
from arquants.helpers import strategy_kwargs_list
from arquants.manager.logger import logger

import arquants.definitions


class Strategy(btStrategy):
    """
    Clase base que presenta la interfaz que deben o pueden implementar las estrategias que extiendan Strategy """
    _callnext = True

    """Routing key in which the PRM workers listen for incoming events."""
    prm_rk = arquants.definitions.configuration.get_property('RABBITMQ', 'PrmConsumerRoutingKey')

    sheet_queue = None
    sheet_reader = None

    def logs(self, message):
        if aqhandler.public:
            logger.info(message)
        else:
            aqhandler.send(json.dumps(eventFactory.createLogsEvent(message).to_dict()), self.prm_rk)

    def log(self, message):
        self.logs(message)

    @staticmethod
    def do_list_of_list(item):
        if type(item.values).__name__ == 'dict':
            row = list()
            for value in item.values.values():
                row.append(value)
            new_values = list()
            new_values.append(row)
            item.values = new_values

    def sheet_log(self, item):
        if sheet_envs.SHEET_INIT:
            self.do_list_of_list(item)
            sheet_envs.LOGGER_QUEUE.put(item)
        else:
            self.logs('[Warning] Intendo loguear sin tener google sheet configurado')

    def row_values(self, sheet_name=None, row=None):
        if sheet_envs.SHEET_INIT:
            return sheet_envs.GSHEET_READER.row_values(sheet_name, row)
        else:
            self.log("[Warning] No hay una sheet configurada")

    def col_values(self, sheet_name=None, col=None):
        if sheet_envs.SHEET_INIT:
            return sheet_envs.GSHEET_READER.col_values(sheet_name, col)
        else:
            self.log("[Warning] No hay una sheet configurada")

    def get_all_values(self, sheet_name):
        if sheet_envs.SHEET_INIT:
            return sheet_envs.GSHEET_READER.get_all_values(sheet_name)
        else:
            self.log("[Warning] No hay una sheet configurada")

    def get_all_records(self, sheet_name):
        if sheet_envs.SHEET_INIT:
            return sheet_envs.GSHEET_READER.get_all_records(sheet_name)
        else:
            self.log("[Warning] No hay una sheet configurada")

    def _next(self):
        try:
            clock_len = self._clk_update()

            for indicator in self._lineiterators[LineIterator.IndType]:
                indicator._next()

            self._notify()

            if self._ltype == LineIterator.StratType:
                # supporting datas with different lengths
                minperstatus = self._getminperstatus()
                if self._callnext:
                    if minperstatus < 0:
                        self.next()
                    elif minperstatus == 0:
                        self.nextstart()  # only called for the 1st value
                    else:
                        self.prenext()
            else:
                # assume indicators and others operate on same length datas
                # although the above operation can be generalized
                if self._callnext:
                    if clock_len > self._minperiod:
                        self.next()
                    elif clock_len == self._minperiod:
                        self.nextstart()  # only called for the 1st value
                    elif clock_len:
                        self.prenext()

            minperstatus = self._getminperstatus()
            self._next_analyzers(minperstatus)
            self._next_observers(minperstatus)

            self.clear()
        except Exception as e:
            self.on_error()
            raise e

    def _start(self):
        super(Strategy, self)._start()
        self.logs("Initializing Strategy...")
        if eventFactory.execution_type == 'live_trading':
            aqhandler.send(json.dumps(eventFactory.createLiveTradingStartedEvent().to_dict()), 'worker')
            logging.getLogger('Strategy')
            logging.info(eventFactory.createLiveTradingStartedEvent().to_dict())
            aqhandler.register_callback(EventType.PARAMETER_CHANGE_EVENT.value, self.process_parameter_change)
            aqhandler.register_callback(EventType.STRATEGY_CHANGE_STATE.value, self.process_status_update)
        self.validate()
        # TODO: utilizar StrategyStatusChengeEvent en lugar de LiveTradingStartedEvent

    def on_pause(self):
        pass

    def on_error(self):
        pass

    def stop(self):
        """
        Averiguar que hace
        :return:
        """
        super(Strategy, self).stop()

    def finish(self):
        """
        Averiguar bien que hace
        :return:
        """
        self.env.runstop()

    def process_parameter_change(self, event):
        """
        Procesa un evento de cambio
        :param event:
        :return:
        """
        for params in event.parameters:
            for name, value in params.items():
                self.__setattr__(name, value)

    def process_status_update(self, event: StrategyStatusChangeEvent):
        """
        Setea _callnext a True o False dependiendo de si el StrategyStatusChangeEvent.status es PAUSED o no.
        :param event:
        :return:
        """
        self._callnext = (event.status != StrategyStatus.PAUSED.value)
        if event.status == StrategyStatus.PAUSED.value:
            self.on_pause()

    def validate(self):
        return True

    def pause(self):
        """
        Pausa la estrategia y publica un evento con su estado
        :return:
        """
        self._callnext = False
        self.on_pause()
        event = eventFactory.createStrategyStatusChangeEvent(status=StrategyStatus.PAUSED.value)
        aqhandler.send(json.dumps(event.to_dict()), routing_key=self.prm_rk)

    def resume(self):
        """
        Resumen al ejecucion de la estrategia y publica un evento con su estado
        :return:
        """
        self._callnext = True
        event = eventFactory.createStrategyStatusChangeEvent(status=StrategyStatus.ACTIVE.value)
        aqhandler.send(json.dumps(event.to_dict()), routing_key=self.prm_rk)

    def __setattr__(self, key, value):
        if key in strategy_kwargs_list.get_kwargs_list():
            value = self._process_parameter(value)
            event = eventFactory.createParameterChangeEvent(parameters={'name': key, 'value': str(value)})
            aqhandler.send(json.dumps(event.to_dict()), self.prm_rk)
        return super().__setattr__(key, value)

    def replace(self, size=None, price=None, order=None, **kwargs):
        """
        Reemplaza una orden
        :param size:
        :param price:
        :param order:
        :param kwargs:
        :return:
        """
        size = size if size is not None else order.size
        price = price if price is not None else order.price

        if size:
            return self.broker.replace(self, size=abs(size), price=price, order=order, **kwargs)

        return None

    def sendOrders(self, orders):
        """  Metodo legacy """
        self.send_orders(orders)

    def send_orders(self, orders):
        self.broker.send_bulk_new_orders(orders)

    @staticmethod
    def _process_parameter(value):
        if isinstance(value, bool) or isinstance(value, float) or isinstance(value, int):
            _value = value
        elif isinstance(value, date):
            _value = value
        elif value == 'True':
            _value = True
        elif value == 'False' or value == 'false':
            _value = False
        else:
            try:
                if "." in value:
                    _value = float(value)
                else:
                    _value = int(value)
            except ValueError:
                _value = value

        return _value

    def alert(self, title=None, body=None):
        event = eventFactory.createPushNotificationsEvent(title=title, body=body)
        aqhandler.send(json.dumps(event.to_dict()), self.prm_rk)

    def mass_cancel(self, routingKey=None, segment=None, account=None):
        self.broker.mass_cancel(routing_key=routingKey, account=account, segment=segment)

    def _notify(self, qorders=[], qtrades=[]):
        try:
            super()._notify(qorders=qorders, qtrades=qtrades)
        except Exception as e:
            self.on_error()
            raise e



import csv
import random
import tempfile
import dateparser

import jsonpickle
from aq_lib.utils import byte_utils
from backtrader.feeds import GenericCSVData
from backtrader.utils import py3 as util
from dateutil.parser import parse


class HAPIData(GenericCSVData):
    TEMP_FILE = tempfile.mkstemp()

    params = (
        ('dataname', TEMP_FILE[1]),
        ('baseurl', 'http://h-api.primary.com.ar/MHD/TradesOHLC'),
        ('fromhour', "00:00:00"),
        ('tohour', "00:00:00"),

        ('datetime', 0),
        ('last_px', 1),
        ('high', 2),
        ('low', 3),
        ('open', 4),
        ('close', 5),
        ('volume', 6),
        ('bid_px', 7),
        ('bid_qty', 8),
        ('offer_px', 9),
        ('offer_qty', 10),
        ('last_qty', 11),
        ('trade_qty', 12),
        ('trade_px', 13),
        ('bid_px_2', 14),
        ('bid_qty_2', 15),
        ('offer_px_2', 16),
        ('offer_qty_2', 17),
        ('bid_px_3', 18),
        ('bid_qty_3', 19),
        ('offer_px_3', 20),
        ('offer_qty_3', 21),
        ('bid_px_4', 22),
        ('bid_qty_4', 23),
        ('offer_px_4', 24),
        ('offer_qty_4', 25),
        ('bid_px_5', 26),
        ('bid_qty_5', 27),
        ('offer_px_5', 28),
        ('offer_qty_5', 29),
    )

    lines = (('bid_px'), ('bid_qty'), ('offer_px'), ('offer_qty'), ('last_px'), ('last_qty'),('trade_px'), ('trade_qty'), ('volume'), ('bid_px_2'), ('bid_qty_2'), ('offer_px_2'), ('offer_qty_2'), ('bid_px_3'), ('bid_qty_3'),
    ('offer_px_3'), ('offer_qty_3'), ('bid_px_4'), ('bid_qty_4'), ('offer_px_4'), ('offer_qty_4'), ('bid_px_5'),
    ('bid_qty_5'), ('offer_px_5'), ('offer_qty_5'))
    _params = params

    def __init__(self, dataname, fromdate, todate):
        self.dataname = dataname
        self.tradecontract = dataname
        self.fieldsnames = [self._params[i][0] for i in range(4, len(self._params))]

        self.fromdate = dateparser.parse(fromdate, date_formats=['%Y-%m-%d'])
        self.todate = dateparser.parse(todate, date_formats=['%Y-%m-%d'])

        url = self.params.baseurl
        url += '/%s' % util.urlquote(dataname)
        # se saca solo el dia desde el string ya que viene en formato yyyy-mm-dd hh:mm:ss
        url += '/%s' % util.urlquote(str(self.fromdate).split(' ')[0])
        url += '/%s' % util.urlquote(str(self.todate).split(' ')[0])
        url += '/%s' % self.params.fromhour
        url += '/%s' % self.params.tohour

        datafile = util.urlopen(url)

        data = datafile.read()
        decodedata = jsonpickle.decode(byte_utils.is_byte_decode(data))
        self.csv_generator(decodedata)

    def csv_generator(self, data):
        with open(self._params[0][1], 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=self.fieldsnames)
            writer.writeheader()
            if data['status'] == 'ERROR':
                raise Exception("No data for this contract")
            else:
                for i in range(len(data['marketDataH'])):
                    partial_line = data['marketDataH'][i]
                    line = self.make_lines(partial_line)
                    writer.writerow(line)

    def make_lines(self, partial_line):
        result_dic = dict()
        for field in self.fieldsnames:
            result_dic[field] = getattr(self, '_' + field)(partial_line)

        return result_dic

    @staticmethod
    def _datetime(line):
        date = line['datetime'].split('.')[0]
        return parse(date)

    @staticmethod
    def _last_px(line):
        return line['price']

    @staticmethod
    def _last_qty(line):
        return line['size']

    @staticmethod
    def _trade_px(line):
        return line['price']

    @staticmethod
    def _trade_qty(line):
        return line['size']

    @staticmethod
    def _high(line):
        return str(float(line['price']) + 1)

    @staticmethod
    def _low(line):
        return str(float(line['price']) - 1)

    @staticmethod
    def _open(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['price']) - random.randint(1, 3))
        return str(float(line['price']) + random.randint(1, 3))

    @staticmethod
    def _close(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['price']) - random.randint(1, 3))
        return str(float(line['price']) + random.randint(1, 3))

    @staticmethod
    def _volume(line):
        return random.randint(22, 100)

    @staticmethod
    def _bid_px(line):
        return str(float(line['price']) - random.randint(1, 3))

    @staticmethod
    def _bid_px_2(line):
        return str(float(line['price']) - (1 + random.randint(1, 3)))

    @staticmethod
    def _bid_px_3(line):
        return str(float(line['price']) - (2 + random.randint(1, 3)))

    @staticmethod
    def _bid_px_4(line):
        return str(float(line['price']) - (3 + random.randint(1, 3)))

    @staticmethod
    def _bid_px_5(line):
        return str(float(line['price']) - (4 + random.randint(1, 3)))

    @staticmethod
    def _bid_qty(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _bid_qty_2(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _bid_qty_3(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _bid_qty_4(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _bid_qty_5(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _offer_px(line):
        return str(float(line['price']) + random.randint(1, 3))

    @staticmethod
    def _offer_px_2(line):
        return str(float(line['price']) + (1 + random.randint(1, 3)))

    @staticmethod
    def _offer_px_3(line):
        return str(float(line['price']) + (2 + random.randint(1, 3)))

    @staticmethod
    def _offer_px_4(line):
        return str(float(line['price']) + (3 + random.randint(1, 3)))

    @staticmethod
    def _offer_px_5(line):
        return str(float(line['price']) + (4 + random.randint(1, 3)))

    @staticmethod
    def _offer_qty(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _offer_qty_2(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _offer_qty_3(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _offer_qty_4(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))

    @staticmethod
    def _offer_qty_5(line):
        signo = random.random()
        if signo < 0.5:
            return str(float(line['size']) - random.randint(1, 3))
        return str(float(line['size']) + random.randint(1, 3))



import datetime
from backtrader import Cerebro, num2date


class AQCerebro(Cerebro):

    def __init__(self, queue):
        super().__init__()
        self.queue = queue

    def _runnext(self, runstrats):
        """
        Actual implementation of run in full next mode. All objects have its
        ``next`` method invoke on each data arrival
        """
        datas = sorted(self.datas,
                       key=lambda x: (x._timeframe, x._compression))
        datas1 = datas[1:]
        data0 = datas[0]
        d0ret = True

        rs = [i for i, x in enumerate(datas) if x.resampling]
        rp = [i for i, x in enumerate(datas) if x.replaying]
        rsonly = [i for i, x in enumerate(datas)
                  if x.resampling and not x.replaying]
        onlyresample = len(datas) == len(rsonly)
        noresample = not rsonly

        clonecount = sum(d._clone for d in datas)
        ldatas = len(datas)
        ldatas_noclones = ldatas - clonecount
        lastqcheck = False
        while d0ret or d0ret is None:

            # señal para activar
            self.queue.get(block=True)

            # if any has live data in the buffer, no data will wait anything
            newqcheck = not any(d.haslivedata() for d in datas)
            if not newqcheck:
                # If no data has reached the live status or all, wait for
                # the next incoming data
                livecount = sum(d._laststatus == d.LIVE for d in datas)
                newqcheck = not livecount or livecount == ldatas_noclones

            lastret = False
            # Notify anything from the store even before moving datas
            # because datas may not move due to an error reported by the store
            self._storenotify()
            if self._event_stop:  # stop if requested
                return
            self._datanotify()
            if self._event_stop:  # stop if requested
                return

            # record starting time and tell feeds to discount the elapsed time
            # from the qcheck value
            drets = []
            qstart = datetime.datetime.utcnow()

            for d in datas:
                qlapse = datetime.datetime.utcnow() - qstart
                d.do_qcheck(newqcheck, qlapse.total_seconds())
                drets.append(d.next(ticks=False))

            d0ret = any((dret for dret in drets))
            if not d0ret and any((dret is None for dret in drets)):
                d0ret = None

            if d0ret:
                dts = []
                for i, ret in enumerate(drets):
                    dts.append(datas[i].datetime[0] if ret else None)

                # Get index to minimum datetime
                if onlyresample or noresample:
                    dt0 = min((d for d in dts if d is not None))
                else:
                    dt0 = min((d for i, d in enumerate(dts)
                               if d is not None and i not in rsonly))

                dmaster = datas[dts.index(dt0)]  # and timemaster
                self._dtmaster = dmaster.num2date(dt0)
                self._udtmaster = num2date(dt0)

                # slen = len(runstrats[0])
                # Try to get something for those that didn't return
                for i, ret in enumerate(drets):
                    if ret:  # dts already contains a valid datetime for this i
                        continue

                    # try to get a data by checking with a master
                    d = datas[i]
                    d._check(forcedata=dmaster)  # check to force output
                    if d.next(datamaster=dmaster, ticks=False):  # retry
                        dts[i] = d.datetime[0]  # good -> store
                        # self._plotfillers2[i].append(slen)  # mark as fill
                    else:
                        # self._plotfillers[i].append(slen)  # mark as empty
                        pass

                # make sure only those at dmaster level end up delivering
                for i, dti in enumerate(dts):
                    if dti is not None:
                        di = datas[i]
                        rpi = False and di.replaying   # to check behavior
                        if dti > dt0:
                            if not rpi:  # must see all ticks ...
                                di.rewind()  # cannot deliver yet
                            # self._plotfillers[i].append(slen)
                        elif not di.replaying:
                            # Replay forces tick fill, else force here
                            di._tick_fill(force=True)

                        # self._plotfillers2[i].append(slen)  # mark as fill

            elif d0ret is None:
                # meant for things like live feeds which may not produce a bar
                # at the moment but need the loop to run for notifications and
                # getting resample and others to produce timely bars
                for data in datas:
                    data._check()
            else:
                lastret = data0._last()
                for data in datas1:
                    lastret += data._last(datamaster=data0)

                if not lastret:
                    # Only go extra round if something was changed by "lasts"
                    break

            # Datas may have generated a new notification after next
            self._datanotify()
            if self._event_stop:  # stop if requested
                return

            if d0ret or lastret:  # if any bar, check timers before broker
                self._check_timers(runstrats, dt0, cheat=True)
                if self.p.cheat_on_open:
                    for strat in runstrats:
                        strat._next_open()
                        if self._event_stop:  # stop if requested
                            return

            self._brokernotify()
            if self._event_stop:  # stop if requested
                return

            if d0ret or lastret:  # bars produced by data or filters
                self._check_timers(runstrats, dt0, cheat=False)
                for strat in runstrats:
                    strat._next()
                    if self._event_stop:  # stop if requested
                        return

                    self._next_writers(runstrats)

        # Last notification chance before stopping
        self._datanotify()
        if self._event_stop:  # stop if requested
            return
        self._storenotify()
        if self._event_stop:  # stop if requested
            return



import math
import uuid

from aq_lib.event.event import *
from aq_lib.event.event_type import EventType
from aq_lib.model.strategy_status import StrategyStatus

from arquants.brokers.bt_order_status_translator import btstatustranslator


class EventFactory:
    def __init__(self, strategy_id=None, strategy_version=None, execution_id=None, arquants_user=None,
                 execution_type=None, routing_key=None):
        self.strategy_id = strategy_id
        self.strategy_version = strategy_version
        self.execution_id = execution_id
        self.arquants_user = arquants_user
        self.execution_type = execution_type
        self.routing_key = routing_key

    # Este evento se crea para backtest, y en live trading solo cuando se envia la orden al mercado
    def createOrderUpdatedEventFromOrder(self, order):
        event = OrderUpdatedEvent(strategy_id=self.strategy_id,
                                  strategy_version=self.strategy_version,
                                  execution_id=self.execution_id,
                                  arquants_user=self.arquants_user,
                                  execution_type=self.execution_type)
        event.account = order.data.account if hasattr(order.data, 'account') else "BT001"
        event.symbol = order.data.tradecontract
        event.side = str(order.ordtype + 1)
        event.orderStatus = btstatustranslator.get(order.getstatusname(order.status), "")
        event.price = order.price if not math.isnan(order.price) else None
        event.orderQty = abs(order.size)
        event.cumQty = 0 if event.orderStatus != "2" else abs(order.size)
        # ref for backtesting, "none" for live trading because it's not defined yet
        event.orderID = order.serverOrderId if hasattr(order, 'serverOrderId') else None
        event.clientOrderID = order.m_orderId if hasattr(order, 'm_orderId') else None
        event.transactTime = str(datetime.now())
        return event

    def createBatchOrderUpdatedEvent(self, order_updated_events):
        order_updated_event = OrderUpdatedEvent(strategy_id=self.strategy_id, strategy_version=self.strategy_version,
                                                execution_id=self.execution_id, arquants_user=self.arquants_user,
                                                topic=EventType.LT_ORDER_UPDATED_EVENT.value,
                                                execution_type=self.execution_type)
        ers = list()
        for savedEvent in order_updated_events.values():
            event = dict()
            event["account"] = "BT001"
            event["symbol"] = savedEvent.symbol
            event["side"] = savedEvent.side
            event["orderStatus"] = savedEvent.orderStatus
            event["price"] = savedEvent.price if not math.isnan(savedEvent.price) else None
            event["orderQty"] = savedEvent.orderQty
            event["cumQty"] = savedEvent.cumQty
            event["orderID"] = savedEvent.orderID
            event["clientOrderID"] = savedEvent.clientOrderID
            event["transactTime"] = savedEvent.transactTime
            ers.append(event)
        order_updated_event.execreports = ers
        return order_updated_event

    def createOrderUpdatedEventFromExecReport(self, exec_report, order):
        event = OrderUpdatedEvent(strategy_id=self.strategy_id, strategy_version=self.strategy_version,
                                  execution_id=self.execution_id, arquants_user=self.arquants_user,
                                  topic=EventType.LT_ORDER_UPDATED_EVENT.value, execution_type=self.execution_type)

        event.account = exec_report.account if exec_report.account is not None else order.data.account
        event.clientOrderID = exec_report.originalClientOrderId if exec_report.originalClientOrderId is not None and exec_report.execType == '4' else exec_report.clientOrderID
        event.cumQty = exec_report.cumQty
        event.execID = exec_report.execID
        event.execType = exec_report.execType
        event.leavesQty = exec_report.leavesQty if exec_report.leavesQty is not None else abs(order.size)
        event.orderID = exec_report.orderID
        event.orderQty = exec_report.orderQty if exec_report.orderQty is not None else abs(order.size)
        event.orderStatus = exec_report.orderStatus
        event.price = exec_report.price if exec_report.price is not None and not math.isnan(exec_report.price) else order.price
        event.securityExchange = exec_report.securityExchange
        event.side = exec_report.side if exec_report.side is not None else str(order.ordtype + 1)
        event.symbol = exec_report.symbol if exec_report.symbol is not None else order.data.tradecontract
        event.text = exec_report.text
        event.timeInForce = exec_report.timeInForce
        event.transactTime = str(exec_report.transactTime)
        return event

    def createBulkNewOrderEvent(self, orders):
        event = BulkNewOrderEvent(EventType.BULK_NEW_ORDER_EVENT.value)
        _orders = []
        for order in orders:
            _orders.append(self.createNewOrderEventFromOrder(order))
        event.orders = _orders
        return event

    def createNewOrderEventFromOrder(self, order):
        event = NewOrderEvent(EventType.NEW_ORDER_EVENT.value)
        event.account = order.data.account
        event.cl_ord_id = order.m_orderId
        event.order_qty = order.abs_size
        event.order_type = str(order.exectype)
        event.price = order.price if order.price is not None and not math.isnan(order.price) else None
        event.side = str(order.ordtype + 1)
        event.symbol = order.data.tradecontract
        event.time_in_force = '0'
        event.routing_key = self.routing_key
        event.settlement_date = order.data.settlement_date
        event.settlement_type = str(order.data.settlement)
        event.security_type = order.data.security_type
        event.currency = order.data.currency
        event.execution_id = self.execution_id
        event.negotiation_agent_id = order.data.negotiation_agent_id
        # if hasattr(order, 'm_short_sell'):
        # event.short_sell = order.short_sell
        if hasattr(order, 'm_execinst'):
            event.exec_inst = order.m_execinst
        return event

    def createCancelOrderEventFromOrder(self, order):
        event = OrderCancelRequestEvent(EventType.ORDER_CANCEL_REQUEST_EVENT.value)
        event.account = order.data.account
        event.cl_ord_id = self.generate_id()
        event.order_id = order.m_serverOrderId  # En las replace no tiene el server_order_id
        event.order_qty = abs(order.size)
        event.orig_cl_ord_id = order.m_orderId
        event.side = str(order.ordtype + 1)
        event.symbol = order.data.tradecontract
        event.routing_key = self.routing_key
        event.settlement_date = order.data.settlement_date
        event.settlement_type = str(order.data.settlement)
        event.security_type = order.data.security_type
        event.currency = order.data.currency
        event.execution_id = self.execution_id
        event.negotiation_agent_id = order.data.negotiation_agent_id
        # if hasattr(order, 'm_short_sell'):
        # event.short_sell = order.short_sell
        return event

    def createMarketDataRequestEvent(self, contract=None, currency=None, md_request_id=None, data_type=None,
                                     settlement=None, settlement_date=None, security_type=None,
                                     negotiation_agent_id=None, market=None):
        return MarketDataRequestEvent(topic=EventType.MARKET_DATA_REQUEST_EVENT.value, symbols=[contract],
                                      md_req_id=md_request_id, currency=currency, routing_key=self.routing_key,
                                      settlement=str(settlement), settlement_date=settlement_date,
                                      security_type=security_type, md_type=data_type, execution_id=self.execution_id,
                                      negotiation_agent_id=negotiation_agent_id,market=market)

    def createExecutionErrorEvent(self, error):
        event = BuildStrategyErrorEvent(execution_id=self.execution_id, arquants_user=self.arquants_user,
                                        strategy_id=self.strategy_id, strategy_version=self.strategy_version,
                                        execution_type=self.execution_type)
        event.error = error
        return event

    # TODO este evento deberia boletearse
    #@staticmethod
    #def createInfoResponseEvent(connectors_rks):
        #    event = InfoResponseEvent()
        #event.connectors_rks = connectors_rks
        #return event

    def createLogsEvent(self, logline):
        event = LogsEvent(arquants_user=self.arquants_user, execution_id=self.execution_id,
                          execution_type=self.execution_type, strategy_id=self.strategy_id)
        event.log_line = logline
        return event

    def createPushNotificationsEvent(self, body=None, title=None):
        event = PushNotificationEvent(body=body, title=title, user=self.arquants_user, execution_id=self.execution_id,
                                      strategy_id=self.strategy_id)
        return event

    # def createLiveTradingEvent(self, datas_positions):
    #     event = LiveTradingEvent(strategy_id=self.strategy_id, strategy_version=self.strategy_version,
    #                              execution_id=self.execution_id, arquants_user=self.arquants_user,
    #                              execution_type=self.execution_type)
    #     event.positions = datas_positions
    #     total_open = 0
    #     total_close = 0
    #     for position in datas_positions.values():
    #         total_open += position.get('open_position')
    #         total_close += position.get('close_position')
    #     event.total_close = total_close
    #     event.total_open = total_open
    #     return event

    def createPositionUpdatedEvent(self, data):
        event = PositionUpdatedEvent(strategy_id=self.strategy_id, strategy_version=self.strategy_version,
                                     execution_id=self.execution_id, arquants_user=self.arquants_user,
                                     execution_type=self.execution_type)
        event.symbol = data["symbol"]
        event.size = data["size"]
        event.price = data["price"]
        event.open_result = data["open_position"]
        event.close_result = data["close_position"]

        return event

    @staticmethod
    def generate_id():
        cl_ord_id = uuid.uuid4().hex[:16]
        return cl_ord_id

    def createTradeExecutedEvent(self, execReport):
        event = TradeExecutedEvent(strategy_id=self.strategy_id, strategy_version=self.strategy_version,
                                   execution_id=self.execution_id, arquants_user=self.arquants_user,
                                   execution_type=self.execution_type)

        event.ts = execReport.transactTime
        event.symbol = execReport.symbol
        event.side = execReport.side
        event.size = execReport.lastQty
        event.price = execReport.lastPrice if not math.isnan(execReport.lastPrice) else None
        event.account = execReport.account
        event.orderid = execReport.orderID
        event.tradeid = execReport.execID
        event.settlType = str(execReport.settlType)
        return event

    def createParameterChangeEvent(self, parameters):
        event = ParameterChangeEvent(parameters=parameters)
        event.execution_id = self.execution_id
        event.arquants_user = self.arquants_user
        event.strategy_id = self.strategy_id
        event.strategy_version = self.strategy_version
        event.execution_type = self.execution_type
        event.from_api = False
        return event

    def createStrategyStatusChangeEvent(self, status: StrategyStatus):
        event = StrategyStatusChangeEvent(status=status, from_api=False, execution_id=self.execution_id)
        return event

    def createLiveTradingStartedEvent(self):
        return LiveTradingStartedEvent(execution_id=self.execution_id, arquants_user=self.arquants_user,
                                        status='active')

    def createOrderCancelReplaceEvent(self, order):
        event = OrderCancelReplaceRequestEvent()

        event.account = order.data.account
        event.cl_ord_id = order.m_orderId
        event.orig_order_id = order.m_originalOrderId
        event.order_qty = abs(order.size)
        event.order_type = str(order.exectype)
        event.orig_client_ord_id = order.m_originalClientOrderId
        event.price = order.price
        event.side = str(order.ordtype + 1)
        event.symbol = order.data.tradecontract
        event.time_in_force = '0'
        event.routing_key = self.routing_key

        event.settlement_type = str(order.data.settlement)
        event.settlement_date = order.data.settlement_date
        event.security_type = order.data.security_type
        event.currency = order.data.currency
        event.execution_id = self.execution_id
        event.negotiation_agent_id = order.data.negotiation_agent_id
        # if hasattr(order, 'm_short_sell'):
        # event.short_sell = order.short_sell
        return event

    # TODO revisar esto
    def createOrderMassCancelRequestEvent(self, segment=None, account=None, clientOrderId=None):
        return OrderMassCancelRequestEvent(topic=EventType.ORDER_MASS_CANCEL_REQUEST_EVENT.value, account=account,
                                           segment=segment, clientOrderId=clientOrderId)


eventFactory = EventFactory()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import datetime
import math
from aq_lib.model.md_entry_type import MDEntryType
from backtrader import date2num
from arquants.feed.base_data import BaseData


class AggregatedData(BaseData):

    data_type = 'aggregated'

    lines = (('bid_px'), ('bid_qty'), ('offer_px'), ('offer_qty'), ('last_px'), ('last_qty'),('volume'),('trade_px'),
             ('trade_qty'),('bid_px_2'), ('bid_qty_2'), ('offer_px_2'), ('offer_qty_2'),('bid_px_3'), ('bid_qty_3'),
             ('offer_px_3'), ('offer_qty_3'),('bid_px_4'), ('bid_qty_4'), ('offer_px_4'), ('offer_qty_4'),('bid_px_5'),
             ('bid_qty_5'), ('offer_px_5'), ('offer_qty_5'),('dummy_line'),('nominal_volume'),('effective_volume'),('close_px'))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.bidPxDepth = []
        self.bidQtyDepth = []
        self.offerPxDepth = []
        self.offerQtyDepth = []
        self.initLinesDept()

    def initLinesDept(self):
        self.bidPxDepth.append(self.lines.bid_px)
        self.bidPxDepth.append(self.lines.bid_px_2)
        self.bidPxDepth.append(self.lines.bid_px_3)
        self.bidPxDepth.append(self.lines.bid_px_4)
        self.bidPxDepth.append(self.lines.bid_px_5)

        self.bidQtyDepth.append(self.lines.bid_qty)
        self.bidQtyDepth.append(self.lines.bid_qty_2)
        self.bidQtyDepth.append(self.lines.bid_qty_3)
        self.bidQtyDepth.append(self.lines.bid_qty_4)
        self.bidQtyDepth.append(self.lines.bid_qty_5)

        self.offerPxDepth.append(self.lines.offer_px)
        self.offerPxDepth.append(self.lines.offer_px_2)
        self.offerPxDepth.append(self.lines.offer_px_3)
        self.offerPxDepth.append(self.lines.offer_px_4)
        self.offerPxDepth.append(self.lines.offer_px_5)

        self.offerQtyDepth.append(self.lines.offer_qty)
        self.offerQtyDepth.append(self.lines.offer_qty_2)
        self.offerQtyDepth.append(self.lines.offer_qty_3)
        self.offerQtyDepth.append(self.lines.offer_qty_4)
        self.offerQtyDepth.append(self.lines.offer_qty_5)

    def _load_md(self, md):
        dt = date2num(datetime.datetime.now())
        # logger.debug("load_md will add data %s" % md.to_dict())
        self.lines.datetime[0] = dt
        loaded = False
        for entry in md.entries:
            position = 0
            if entry.position is not None:
                position = entry.position - 1
            if entry.entry_type == MDEntryType.BID.value:
                if position <= 4:
                    self.bidPxDepth[position][0] = entry.entry_px
                    self.bidQtyDepth[position][0] = entry.entry_size
                    loaded = True
            elif entry.entry_type == MDEntryType.OFFER.value:
                if position <= 4:
                    self.offerPxDepth[position][0] = entry.entry_px
                    self.offerQtyDepth[position][0] = entry.entry_size
                    loaded = True
            elif entry.entry_type == MDEntryType.TRADE.value:
                if entry.trd_type is not None:
                    self.lines.trade_px[0] = entry.entry_px
                    self.lines.trade_qty[0] = entry.entry_size
                    # Acumulo el volumen operado visto
                    if not math.isnan(self.lines.volume[-1]):
                        self.lines.volume[0] = self.lines.volume[-1] + entry.entry_size
                    else:
                        self.lines.volume[0] = entry.entry_size
                self.lines.last_px[0] = entry.entry_px
                self.lines.last_qty[0] = entry.entry_size
                self.position.update_on_md(self.tradecontract,entry.entry_px)
                loaded = True
            elif entry.entry_type == MDEntryType.CLOSE.value:
                self.lines.close_px[0] = entry.entry_px
            elif entry.entry_type == MDEntryType.TRADE_VOLUME.value:
                self.lines.effective_volume[0] = entry.entry_px
                self.lines.nominal_volume[0] = entry.entry_size
                loaded = True
        # Intentamos trasladar el volumen del momento anterior al actual para no perder el acumulado
        if math.isnan(self.lines.volume[0]):
            if not math.isnan(self.lines.volume[-1]):
                self.lines.volume[0] = self.lines.volume[-1]
                loaded = True
        if not loaded:
            self.lines.dummy_line[0] = 0
        return True



import json
from oauth2client.service_account import ServiceAccountCredentials
from arquants.sheets.gsheet_reader import SheetReader
from arquants.sheets.queue_reader import QueueReader
from arquants.sheets.accumulator import Accumulator
from arquants.sheets.batch_logger import BatchLogger


class SheetManager:

    scope = ["https://spreadsheets.google.com/feeds", 'https://www.googleapis.com/auth/spreadsheets',
             "https://www.googleapis.com/auth/drive.file", "https://www.googleapis.com/auth/drive"]

    def __init__(self, google_json=None, spreadsheet_id=None, queue=None):
        self.logger_queue = queue
        self.spreadsheet_id = spreadsheet_id
        print(google_json)
        google_dict = json.loads(google_json)
        self.credentials = ServiceAccountCredentials.from_json_keyfile_dict(google_dict, self.scope)

        self.batch_logger = BatchLogger(self.credentials, self.spreadsheet_id)
        self.accumulator = Accumulator(self.batch_logger.process_items)
        self.queue_reader = QueueReader(queue=self.logger_queue, processor=self.accumulator)

        self.queue_reader.start()
        self.accumulator.start()
        self.gsheet_reader = SheetReader(self.credentials, self.spreadsheet_id)


#--------------------------Test Code--------------------------
# class SheetItemMock:
#
#     def __init__(self,worksheet_name=None,cell=None,values=[]):
#         self.worksheet_name = worksheet_name
#         self.cell = cell
#         self.values = values
#
# google_json = '{"type": "service_account", "project_id": "quantsheets","private_key_id": "81c822a44a4eaa13bebb1cce0f5616d78e5cbdee",' \
#         '"private_key": "-----BEGIN PRIVATE KEY-----\\nMIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQC4H3miSsZgkFgZ\\n0n+mN45gyky8QaJoaZg2B/UwDxjcdY3+f9EyYqWCdnmCT0NL2yGHJ+jUeH3P48CN\\n/n87Xk24LLbBaDawFvquewlyYBIiPI03fy1U04SvkkJ+P/n1eXeFOWrJc6RYF7eH\\nutJ42nD8WfhKdgHYA3O3ckUO55LENztJQefongG3la4jyo9HTGCO/i5k6WqSqzTF\\nY9yDhNVTGgFwEmGDXAZXxh0J8AJzmMpnBtJ9hR6NQZDKQrORHcFdoQBkMe/RLlOm\\n/8n37ZebEqHkXftyRUGpSJbItswBxEkSv53SnCFnFEpd5oSfzthJaIwZtRV9Bf6W\\nQ83VLnAlAgMBAAECggEACTUuapwQMHnkz1fG9puZJRnWKA+lN6ZY4bzYpBBhNycx\\nfxcfxwQGuw/siC8pZ6SmrYEPWaQemz6hyB1rJJ9/m1SwcrAg6gH1g95sd286q5fO\\ncnEg+yNqeN2Y3IJGYGJfGdxoTM5EcKZbpoVEgMg1yxzpt+Q6Yosw+9fSVco6vVFg\\nVJ4rYeleP9Thh8PUVlN9thpNzBvmm1Reb0cFrr3va519KZyNUOmtoSSEngCEU7sA\\nsetJPpk3uQ5jkfBDA2/xfjOhxEkwmmIkKtnIR5q20UM3i4yDU/nY5BWBhAwivatR\\n8VRZsx7drS089dqxIKBZLgne3ir+PyZwbV5oSaKXOQKBgQDq6S8rDW6cqtNdxmaC\\nknlDz3hEyERmntFbUI2pwkojDStV36o1lr9sVq2ZSN+wqYaUuqR2c0QGhVYC1TE3\\ngOp0A9DW7W9sJqz+ZVLrcVYuRD6ZveizEoHFze9JiZT5TsE3/VITv9/BfCfcE5Yt\\nbz/INtn1xg2rOQIjeOimXJX0zQKBgQDIpw/7+4XTACJQ1chjKYxBvt1OZof+XoMy\\nrfH54gUX8Vvkaf4Qe/h5u5DqFxDT2YYz6LMqNUvGwVu3sddrfQSXVFlC0tlHdPEG\\n3EdtI1H76juTV29tc4gGt14ZBOpDnCPy7Y5D6bJfdW0iA86xIobjrPUj7Lp8uJfb\\nGofuhNKouQKBgDDToyA5vIIH58MYF/qP43C60O7LCZyi6jUmuytL3QkIFbfVs5VL\\n7iHgqliwEv6vXe/QE5sjPkJ0uHoCQiadPx46JNBnrb83EsIV4XRarGUVfkWKebGu\\n6RZqRZBtbm+bdQHkP4knWqTm34oY8CAlfYZqEEfLkM/EG7Ovz/u0Rt8JAoGATsbY\\nTbPoTHjnABmOvO/Y8w7+UAONLN7qX4FPWS3VlhraWwkKCGrDmPd844r+vk3OlJ6t\\niMq332aWnb4itz60CL3C9atWRumwn2LoX/7X9zF1Blnzk2MiapCfTu9REg5BRuTh\\nT+R2dWfi2tHn2j+V+dkzvVD5vnGynQEDkS9wjXECgYEA45OQDUYxTEuVr4ZUKVvU\\nYC2fIKeetiqo04yia48I6OZ8J7lNIxPL4l2YwGN95I1iSxwbSbUXk7E0227ygZST\\nWsOLOFktfnbSKBulSgPRiqXyI469M1Rc6qccoprfiy4Adw3Cd1pwX5KOpZEhzVVT\\nPS9EgZJ2ppKPgFeZ+X6t1kw=\\n-----END PRIVATE KEY-----\\n",' \
#         '"client_email": "quantsheet@quantsheets.iam.gserviceaccount.com",' \
#         '"client_id": "113805095232193113831",' \
#         '"auth_uri": "https://accounts.google.com/o/oauth2/auth",' \
#         '"token_uri": "https://oauth2.googleapis.com/token",' \
#         '"auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",' \
#         '"client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/quantsheet%40quantsheets.iam.gserviceaccount.com"}'
#
# sl = SheetManager(spreadsheet_id='Test', google_json=google_json)
#
# a=dict()
# a['columna1'] = 'raul'
# a['columna2'] = 'pepito'
# a['columna3'] = 'cachito'
# item = SheetItemMock(worksheet_name='asdasd', cell='B2')
# item.values = a
#
# sl.accumulator.process_item(item)
#
# result = sl.gsheet_reader.get_all_records(sheet_name='asdasd')
# print(result)



import datetime

import tempfile

import random
from backtrader.feeds import GenericCSVData
import csv


class MockData(GenericCSVData):
    TEMP_FILE = tempfile.mkstemp()

    _days_offset = None
    _LAST_PX = None

    params = (
        ('dataname', TEMP_FILE[1]),
        ('fromdate', datetime.datetime(2000, 1, 1)),
        ('todate', datetime.datetime(2000, 2, 27)),
        ('datetime', 0),
        ('last_px', 1),
        ('high', 2),
        ('low', 3),
        ('open', 4),
        ('close', 5),
        ('volume', 6),
        ('bid_px', 7),
        ('bid_qty', 8),
        ('offer_px', 9),
        ('offer_qty', 10),
        ('last_qty', 11),
        ('trade_px', 12),
        ('trade_qty', 13),
        ('bid_px_2', 14),
        ('bid_qty_2', 15),
        ('offer_px_2', 16),
        ('offer_qty_2', 17),
        ('bid_px_3', 18),
        ('bid_qty_3', 19),
        ('offer_px_3', 20),
        ('offer_qty_3', 21),
        ('bid_px_4', 22),
        ('bid_qty_4', 23),
        ('offer_px_4', 24),
        ('offer_qty_4', 25),
        ('bid_px_5', 26),
        ('bid_qty_5', 27),
        ('offer_px_5', 28),
        ('offer_qty_5', 29),
    )

    lines = (
    ('bid_px'), ('bid_qty'), ('offer_px'), ('offer_qty'), ('last_px'), ('last_qty'), ('trade_px'), ('trade_qty'),
    ('volume'), ('bid_px_2'), ('bid_qty_2'), ('offer_px_2'), ('offer_qty_2'), ('bid_px_3'), ('bid_qty_3'),
    ('offer_px_3'), ('offer_qty_3'), ('bid_px_4'), ('bid_qty_4'), ('offer_px_4'), ('offer_qty_4'), ('bid_px_5'),
    ('bid_qty_5'), ('offer_px_5'), ('offer_qty_5'))
    _params = params

    def __init__(self, tradecontract, *args, **kwargs):
        self.fieldsnames = [self._params[i][0] for i in range(3, len(self._params))]
        self.csv_generator()
        self.tradecontract = 'data' + str(tradecontract)
        super().__init__(*args, **kwargs)

    def csv_generator(self):
        with open(self._params[0][1], 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=self.fieldsnames)
            writer.writeheader()
            for day in range(self.get_total_days()):
                line = self.make_lines(day)
                writer.writerow(line)

    def get_total_days(self):
        return (self._params[2][1] - self._params[1][1]).days

    def make_lines(self, day):
        self._days_offset = day
        result_dic = dict()
        for field in self.fieldsnames:
            result_dic[field] = getattr(self, '_' + field)()

        return result_dic

    def _datetime(self):
        return self._params[1][1] - datetime.timedelta(days=-self._days_offset)

    def _last_px(self):
        if self._LAST_PX:
            self._LAST_PX += random.random() * random.randrange(-1, 2)
            return self._LAST_PX
        else:
            self._LAST_PX = random.randint(22, 1509) + random.random()
            return self._LAST_PX

    def _high(self):
        return self._LAST_PX + 3

    def _low(self):
        return self._LAST_PX - 3

    def _open(self):
        return self._LAST_PX - 2

    def _close(self):
        return self._LAST_PX + 2

    @staticmethod
    def _volume():
        return random.randint(22, 100)

    def _bid_px(self):
        return self._LAST_PX - random.random()

    def _bid_px_2(self):
        return self._LAST_PX - (1 + random.random())

    def _bid_px_3(self):
        return self._LAST_PX - (2 + random.random())

    def _bid_px_4(self):
        return self._LAST_PX - (3 + random.random())

    def _bid_px_5(self):
        return self._LAST_PX - (4 + random.random())

    @staticmethod
    def _bid_qty():
        return random.randint(5, 10)

    @staticmethod
    def _bid_qty_2():
        return random.randint(5, 10)

    @staticmethod
    def _bid_qty_3():
        return random.randint(5, 10)

    @staticmethod
    def _bid_qty_4():
        return random.randint(5, 10)

    @staticmethod
    def _bid_qty_5():
        return random.randint(5, 10)

    def _offer_px(self):
        return self._LAST_PX + random.random()

    def _offer_px_2(self):
        return self._LAST_PX + (1 + random.random())

    def _offer_px_3(self):
        return self._LAST_PX + (2 + random.random())

    def _offer_px_4(self):
        return self._LAST_PX + (3 + random.random())

    def _offer_px_5(self):
        return self._LAST_PX + (4 + random.random())

    @staticmethod
    def _offer_qty():
        return random.randint(5, 10)

    @staticmethod
    def _offer_qty_2():
        return random.randint(5, 10)

    @staticmethod
    def _offer_qty_3():
        return random.randint(5, 10)

    @staticmethod
    def _offer_qty_4():
        return random.randint(5, 10)

    @staticmethod
    def _offer_qty_5():
        return random.randint(5, 10)

    @staticmethod
    def _last_qty():
        return random.randint(5, 10)

    @staticmethod
    def _trade_px():
        return random.randint(5, 10)

    @staticmethod
    def _trade_qty():
        return random.randint(5, 10)



from googleapiclient.discovery import build
import traceback
import json
from arquants.sheets.json_generator import RequestGenerator


class BatchLogger:

    scope = ["https://spreadsheets.google.com/feeds", 'https://www.googleapis.com/auth/spreadsheets',
             "https://www.googleapis.com/auth/drive.file", "https://www.googleapis.com/auth/drive"]

    def __init__(self, creds, spreadsheet_id):
        self.api = build('sheets', 'v4', credentials=creds).spreadsheets()
        self.generator = RequestGenerator()
        self.spreadsheet_id = spreadsheet_id
        self.sheet_names = dict()
        self._load_sheet_ids()

    def process_items(self, items):
        print('procesando item')
        requests = list()

        for item in items:
            sheet_id = self.get_or_create_sheet(item.worksheet_name)
            if not item.cell:
                request = self.generator.gen_json_append_request(item.values, sheet_id)
            else:
                request = self.generator.gen_json_update_request(item.values, sheet_id, item.cell)
            requests.append(request)

        self._batch_execute(requests)

    def _batch_execute(self, requests):
        try:
            request_body = {'requests': requests}
            response = self.api.batchUpdate(spreadsheetId=self.spreadsheet_id, body=request_body).execute()
            return response
        except Exception as e:
            log(traceback.format_exc())

    def _load_sheet_ids(self):
        spreadsheet = self.api.get(spreadsheetId=self.spreadsheet_id, fields='sheets(data/rowData/values/userEnteredValue,properties(index,sheetId,title))').execute()
        for sheet in spreadsheet['sheets']:
            sheet_name = sheet['properties']['title']
            sheet_id = sheet['properties']['sheetId']
            self.sheet_names[sheet_name] = sheet_id
        print('Loaded sheets : ', self.sheet_names)
        print('spreadsheet info : ', spreadsheet)
        print(json.dumps(spreadsheet, sort_keys=True, indent=4))

    def get_or_create_sheet(self, name):
        if not self.sheet_names.get(name):
            log('sheet {} does not exist, will create it'.format(name))
            self.sheet_names[name] = self.create_sheet(name)
        return self.sheet_names[name]

    def create_sheet(self, sheet_name):
        try:
            request_body = {
                'requests': [{
                    'addSheet': {
                        'properties': {
                            'title': sheet_name,
                        }
                    }
                }]
            }
            response = self.api.batchUpdate(
                spreadsheetId=self.spreadsheet_id,
                body=request_body
            ).execute()
            sheet_id = response['replies'][0]['addSheet']['properties']['sheetId']
            return sheet_id

        except Exception as e:
            log(traceback.format_exc())

    # Draft implementation =========================================================================
    # ==============================================================================================

    def append(self, sheet_name, values):
        self.ensure_sheet_exists(sheet_name)

        body = {
            'values': values
        }
        result = self.api.values().append(spreadsheetId=self.spreadsheet_id, range=sheet_name,
                                          valueInputOption='USER_ENTERED', body=body).execute()
        print(result)

    def update(self, update_range, values):
        sheet_name = update_range.split(',')[0]
        log('Updating sheet {}'.format(sheet_name))

        self.ensure_sheet_exists(sheet_name)

        body = {
            'values': values
        }
        result = self.api.values().update(spreadsheetId=self.spreadsheet_id, range=update_range,
                                          valueInputOption='USER_ENTERED', body=body).execute()
        print(result)

    def get(self, get_range):
        result = self.api.values().get(spreadsheetId=self.spreadsheet_id, range=get_range).execute()
        print(result)

def log(message):
    print(message)

# ==================================================================================================
# TEST HARNESS
# ==================================================================================================
# class MockItem:

#    def __init__(self,worksheet_name=None,cell=None,values=[]):
#        self.worksheet_name = worksheet_name
#        self.cell = cell
#        self.values = values

# class Glogger_tester:
#    def run(self):

#        creds = ServiceAccountCredentials.from_json_keyfile_name("creds.json", BatchLogger.scope)
#        sid = '1k7Wmi-Wsb-LPdEBhGOI6uiztwTkcsPPBfiUFmCyNRm0'
#        glogger = BatchLogger(creds, sid)

#        append_values = [
#            ['cadillac', 'eldorado', 1973, 30000],
#            ['maserati', 'GT', 80000]]
#         append_item = MockItem()
#         append_item.values = append_values
#         append_item.worksheet_name = 'Cars'
#
#         update_values = [
#             ['Peugeot', '308 SW'],
#             ['Peugeot', '406']]
#         update_item = MockItem()
#         update_item.values = update_values
#         update_item.cell = 'A2'
#         update_item.worksheet_name = 'Cars'

        # glogger.process_items([append_item, update_item])

# glogger_tester = Glogger_tester()
# glogger_tester.run()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import datetime
import math
from aq_lib.model.md_entry_type import MDEntryType
from aq_lib.utils.logger import logger
from backtrader import date2num

from arquants import BaseData


class DisaggregatedData(BaseData):

    data_type = 'disaggregated'

    lines = (('bid_px_1'), ('bid_qty_1'), ('offer_px_1'), ('offer_qty_1'), ('last_px'), ('last_qty') ,('volume'),
             ('trade_px'), ('trade_qty'), ('bid_px_2'), ('bid_qty_2'), ('offer_px_2'), ('offer_qty_2'), ('bid_px_3'),
             ('bid_qty_3'), ('offer_px_3'), ('offer_qty_3'), ('bid_px_4'), ('bid_qty_4'), ('offer_px_4'), ('offer_qty_4'),
             ('bid_px_5'), ('bid_qty_5'), ('offer_px_5'), ('offer_qty_5'), ('dummy_line'), ('nominal_volume'),
             ('effective_volume'), ('close_px'))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.a = 8
        self.bidPxDepth = []
        self.bidQtyDepth = []
        self.offerPxDepth = []
        self.offerQtyDepth = []
        self.bidOrdersDepth = []
        self.offerOrdersDepth = []

        self.bid_orders_1 = list()
        self.bid_orders_2 = list()
        self.bid_orders_3 = list()
        self.bid_orders_4 = list()
        self.bid_orders_5 = list()

        self.offer_orders_1 = list()
        self.offer_orders_2 = list()
        self.offer_orders_3 = list()
        self.offer_orders_4 = list()
        self.offer_orders_5 = list()

        self.init_lines_dept()

    def init_lines_dept(self):

        self.bidPxDepth.append(self.lines.bid_px_1)
        self.bidPxDepth.append(self.lines.bid_px_2)
        self.bidPxDepth.append(self.lines.bid_px_3)
        self.bidPxDepth.append(self.lines.bid_px_4)
        self.bidPxDepth.append(self.lines.bid_px_5)

        self.bidQtyDepth.append(self.lines.bid_qty_1)
        self.bidQtyDepth.append(self.lines.bid_qty_2)
        self.bidQtyDepth.append(self.lines.bid_qty_3)
        self.bidQtyDepth.append(self.lines.bid_qty_4)
        self.bidQtyDepth.append(self.lines.bid_qty_5)

        self.offerPxDepth.append(self.lines.offer_px_1)
        self.offerPxDepth.append(self.lines.offer_px_2)
        self.offerPxDepth.append(self.lines.offer_px_3)
        self.offerPxDepth.append(self.lines.offer_px_4)
        self.offerPxDepth.append(self.lines.offer_px_5)

        self.offerQtyDepth.append(self.lines.offer_qty_1)
        self.offerQtyDepth.append(self.lines.offer_qty_2)
        self.offerQtyDepth.append(self.lines.offer_qty_3)
        self.offerQtyDepth.append(self.lines.offer_qty_4)
        self.offerQtyDepth.append(self.lines.offer_qty_5)

        self.bidOrdersDepth.append(self.bid_orders_1)
        self.bidOrdersDepth.append(self.bid_orders_2)
        self.bidOrdersDepth.append(self.bid_orders_3)
        self.bidOrdersDepth.append(self.bid_orders_4)
        self.bidOrdersDepth.append(self.bid_orders_5)

        self.offerOrdersDepth.append(self.offer_orders_1)
        self.offerOrdersDepth.append(self.offer_orders_2)
        self.offerOrdersDepth.append(self.offer_orders_3)
        self.offerOrdersDepth.append(self.offer_orders_4)
        self.offerOrdersDepth.append(self.offer_orders_5)

    def _load_md(self, md):
        dt = date2num(datetime.datetime.now())
        self.lines.datetime[0] = dt
        loaded = False
        bid_loaded = False
        offer_loaded = False
        bid_position = 0
        offer_position = 0
        bid_last_price = -1
        offer_last_price = -1
        clean_bid = False
        clean_offer = False
        for entry in md.entries:

            if entry.entry_type == MDEntryType.BID.value:
                if not clean_bid:
                    clean_bid = True
                    self.clean_orders(self.bidOrdersDepth)
                if bid_position <= 4:
                    if bid_last_price == -1:
                        bid_last_price = entry.entry_px

                    if bid_last_price == entry.entry_px:
                        self.bidOrdersDepth[bid_position].append(entry.entry_size)
                        if not bid_loaded:
                            self.bidPxDepth[bid_position][0] = entry.entry_px
                            bid_loaded = True
                    else:
                        bid_position += 1
                        if bid_position <= 4:
                            self.bidPxDepth[bid_position][0] = entry.entry_px
                            self.bidOrdersDepth[bid_position].append(entry.entry_size)
                            bid_last_price = entry.entry_px
                            bid_loaded = True

            elif entry.entry_type == MDEntryType.OFFER.value:
                if not clean_offer:
                    self.clean_orders(self.offerOrdersDepth)
                    clean_offer = True
                if offer_position <= 4:

                    if offer_last_price == -1:
                        offer_last_price = entry.entry_px

                    if offer_last_price == entry.entry_px:
                        self.offerOrdersDepth[offer_position].append(entry.entry_size)
                        if not offer_loaded:
                            self.offerPxDepth[offer_position][0] = entry.entry_px
                            offer_loaded = True
                    else:
                        offer_position += 1
                        if offer_position <= 4:
                            self.offerPxDepth[offer_position][0] = entry.entry_px
                            self.offerOrdersDepth[offer_position].append(entry.entry_size)
                            offer_last_price = entry.entry_px
                            offer_loaded = True

            elif entry.entry_type == MDEntryType.TRADE.value:
                if entry.trd_type is not None:
                    self.lines.trade_px[0] = entry.entry_px
                    self.lines.trade_qty[0] = entry.entry_size
                    if not math.isnan(self.lines.volume[-1]):
                        self.lines.volume[0] = self.lines.volume[-1] + entry.entry_size
                    else:
                        self.lines.volume[0] = entry.entry_size
                self.lines.last_px[0] = entry.entry_px
                self.lines.last_qty[0] = entry.entry_size
                logger.debug("saving last price in position for %s" % self.tradecontract)
                self.position.update_on_md(self.tradecontract, entry.entry_px)
                loaded = True
            elif entry.entry_type == MDEntryType.CLOSE.value:
                self.lines.close_px[0] = entry.entry_px
                loaded = True
            elif entry.entry_type == MDEntryType.TRADE_VOLUME.value:
                self.lines.effective_volume[0] = entry.entry_px
                self.lines.nominal_volume[0] = entry.entry_size
                loaded = True

        for i in range(0, 5):
            orders = self.offerOrdersDepth[i]
            if orders:
                self.offerQtyDepth[i] = sum(orders)
            orders = self.bidOrdersDepth[i]
            if orders:
                self.bidQtyDepth[i] = sum(orders)

        if math.isnan(self.lines.volume[0]):
            if not math.isnan(self.lines.volume[-1]):
                self.lines.volume[0] = self.lines.volume[-1]
                loaded = True

        if not (loaded or bid_loaded or offer_loaded):
            self.lines.dummy_line[0] = 0
        return True

    def clean_orders(self, orders_depth):
        for i in range(0, len(orders_depth)):
            orders_depth[i].clear()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import abc
import math
import backtrader as bt
from backtrader import TimeFrame, date2num, num2date
from backtrader.feed import DataBase
from backtrader.utils.py3 import (queue, string_types,
                                  with_metaclass)
from arquants.stores import aq_store
from arquants.stores.aq_store import AQStore


class MetaData(DataBase.__class__):
    def __init__(cls, name, bases, dct):
        """Class has already been created ... register"""
        # Initialize the class
        super(MetaData, cls).__init__(name, bases, dct)

        if 'data_type' in dct:
            data_type = dct.get('data_type')
            if data_type:
                # Register with the store
                aq_store.AQStore.datas_class[data_type] = cls
        else:
            raise Exception('dame un data type')


class BaseData(with_metaclass(MetaData, DataBase)):
    """

    Params:

      - ``sectype`` (default: ``STK``)

        Default value to apply as *security type* if not provided in the
        ``dataname`` specification

      - ``exchange`` (default: ``SMART``)

        Default value to apply as *exchange* if not provided in the
        ``dataname`` specification

      - ``currency`` (default: ``''``)

        Default value to apply as *currency* if not provided in the
        ``dataname`` specification

      - ``historical`` (default: ``False``)

        If set to ``True`` the data feed will stop after doing the first
        download of data.

        The standard data feed parameters ``fromdate`` and ``todate`` will be
        used as reference.

        The data feed will make multiple requests if the requested duration is
        larger than the one allowed by IB given the timeframe/compression
        chosen for the data.

      - ``what`` (default: ``None``)

        If ``None`` the default for different assets types will be used for
        historical data requests:

          - 'BID' for CASH assets
          - 'TRADES' for any other

        Check the IB API docs if another value is wished

      - ``rtbar`` (default: ``False``)

        If ``True`` the ``5 Seconds Realtime bars`` provided by Interactive
        Brokers will be used as the smalles tick. According to the
        documentation they correspond to real-time values (once collated and
        curated by IB)

        If ``False`` then the ``RTVolume`` prices will be used, which are based
        on receiving ticks. In the case of ``CASH`` assets (like for example
        EUR.JPY) ``RTVolume`` will always be used and from it the ``bid`` price
        (industry de-facto standard with IB according to the literature
        scattered over the Internet)

        Even if set to ``True``, if the data is resampled/kept to a
        timeframe/compression below Seconds/5, no real time bars will be used,
        because IB doesn't serve them below that level

      - ``qcheck`` (default: ``0.5``)

        Time in seconds to wake up if no data is received to give a chance to
        resample/replay packets properly and pass notifications up the chain

      - ``backfill_start`` (default: ``True``)

        Perform backfilling at the start. The maximum possible historical data
        will be fetched in a single request.

      - ``backfill`` (default: ``True``)

        Perform backfilling after a disconnection/reconnection cycle. The gap
        duration will be used to download the smallest possible amount of data

      - ``backfill_from`` (default: ``None``)

        An additional data source can be passed to do an initial layer of
        backfilling. Once the data source is depleted and if requested,
        backfilling from IB will take place. This is ideally meant to backfill
        from already stored sources like a file on disk, but not limited to.

      - ``latethrough`` (default: ``False``)

        If the data source is resampled/replayed, some ticks may come in too
        late for the already delivered resampled/replayed bar. If this is
        ``True`` those ticks will bet let through in any case.

        Check the Resampler documentation to see who to take those ticks into
        account.

        This can happen especially if ``timeoffset`` is set to ``False``  in
        the ``IBStore`` instance and the TWS server time is not in sync with
        that of the local computer

      - ``tradename`` (default: ``None``)
        Useful for some specific cases like ``CFD`` in which prices are offered
        by one asset and trading happens in a different onel

        - SPY-STK-SMART-USD -> SP500 ETF (will be specified as ``dataname``)

        - SPY-CFD-SMART-USD -> which is the corresponding CFD which offers not
          price tracking but in this case will be the trading asset (specified
          as ``tradename``)

    The default values in the params are the to allow things like ```TICKER``,
    to which the parameter ``sectype`` (default: ``STK``) and ``exchange``
    (default: ``SMART``) are applied.

    Some assets like ``AAPL`` need full specification including ``currency``
    (default: '') whereas others like ``TWTR`` can be simply passed as it is.

      - ``AAPL-STK-SMART-USD`` would be the full specification for dataname

        Or else: ``IBData`` as ``IBData(dataname='AAPL', currency='USD')``
        which uses the default values (``STK`` and ``SMART``) and overrides
        the currency to be ``USD``
    """

    data_type = None

    params = (
        ('sectype', 'STK'),  # usual industry value
        ('exchange', 'SMART'),  # usual industry value
        ('currency', ''),
        ('rtbar', False),  # use RealTime 5 seconds bars
        ('historical', False),  # only historical download
        ('what', None),  # historical - what to show
        ('useRTH', False),  # historical - download only Regular Trading Hours
        ('qcheck', 0.0001),  # timeout in seconds (float) to check for events
        ('backfill_start', True),  # do backfilling at the start
        ('backfill', True),  # do backfilling when reconnecting
        ('backfill_from', None),  # additional data source to do backfill from
        ('latethrough', False),  # let late samples through
        ('tradename', None),  # use a different asset as order target
        ('contract', None),  # use a different asset as order target
        ('market', None),  # use a different asset as order target
        ('account', None),  # account
        ('routingkey', None),  # targetRoutingKey for this data
        ('settlement', None),  # targetRoutingKey for this data
        ('settlement_date', None),  # targetRoutingKey for this data
        ('security_type', None),  # targetRoutingKey for this data
        ('currency', None),  # targetRoutingKey for this data
        ('price_size', None),  # targetRoutingKey for this data
        ('contractsize', None),
        ('negotiation_agent_id', None),
        ('variacion_rf', None),
        ('instrument_type', None),
        ('variacion_rv', None),
    )

    _store = AQStore

    # Minimum size supported by real-time bars
    RTBAR_MINSIZE = (TimeFrame.Seconds, 5)

    # States for the Finite State Machine in _load
    _ST_FROM, _ST_START, _ST_LIVE, _ST_HISTORBACK, _ST_OVER = range(5)

    def _timeoffset(self):
        return self.ib.timeoffset()

    def _gettz(self):
        # If no object has been provided by the user and a timezone can be
        # found via contractdtails, then try to get it from pytz, which may or
        # may not be available.

        # The timezone specifications returned by TWS seem to be abbreviations
        # understood by pytz, but the full list which TWS may return is not
        # documented and one of the abbreviations may fail
        tzstr = isinstance(self.p.tz, string_types)
        if self.p.tz is not None and not tzstr:
            return bt.utils.date.Localizer(self.p.tz)

        self.contractdetails = None
        if self.contractdetails is None:
            return None  # nothing can be done

        try:
            import pytz  # keep the import very local
        except ImportError:
            return None  # nothing can be done

        tzs = self.p.tz if tzstr else self.contractdetails.m_timeZoneId

        if tzs == 'CST':  # reported by TWS, not compatible with pytz. patch it
            tzs = 'CST6CDT'

        try:
            tz = pytz.timezone(tzs)
        except pytz.UnknownTimeZoneError:
            return None  # nothing can be done

        # contractdetails there, import ok, timezone found, return it
        return tz

    def islive(self):
        """Returns ``True`` to notify ``Cerebro`` that preloading and runonce
        should be deactivated"""
        return not self.p.historical

    def __init__(self, **kwargs):
        self.ib = self._store(**kwargs)
        self.precontract = self.p.dataname
        self.pretradecontract = self.p.tradename
        self.tradecontract = self.p.dataname
        self.contractsize = self.p.contractsize
        self.account = self.p.account
        self.contract = self.p.dataname
        self.market = self.p.market
        self.settlement = self.p.settlement
        self.currency = self.p.currency
        self.price_size = self.p.price_size
        self.routingkey = self.p.routingkey
        self.position = self.ib.getposition(self.tradecontract, False)
        self.variacion_rf = self.p.variacion_rf
        self.variacion_rv = self.p.variacion_rv
        self.instrument_type = self.p.instrument_type
        self.settlement_date = self.p.settlement_date
        self.security_type = self.p.security_type
        self.negotiation_agent_id = self.p.negotiation_agent_id

    def __len__(self):
        try:
            return super().__len__()
        except ValueError:
            return 0

    def setenvironment(self, env):
        """Receives an environment (cerebro) and passes it over to the store it
        belongs to"""
        super(BaseData, self).setenvironment(env)
        env.addstore(self.ib)

    def start(self):
        super(BaseData, self).start()
        self.qlive = self.ib.start(data=self)
        self._state = self._ST_START
        self._storedmsg = dict()  # keep pending live message (under None)
        self.put_notification(self.CONNECTED)
        if self._state == self._ST_START:
            self._start_finish()  # to finish initialization
            self._st_start()

    def stop(self):
        """Stops and tells the store to stop"""
        super(BaseData, self).stop()
        self.ib.stop()

    def reqdata(self):
        """request real-time data. checks cash vs non-cash) and param useRT"""
        if self.contract is None:
            return

        if self._usertvol:
            self.qlive = self.ib.reqMktData(self.contract)
            self.qlive = self.ib.reqMktData(self.contract)
        else:
            self.qlive = self.ib.reqRealTimeBars(self.contract)

        return self.qlive

    def canceldata(self):
        """Cancels Market Data subscription, checking asset type and rtbar"""
        if self.contract is None:
            return

        if self._usertvol:
            self.ib.cancel_mkt_data(self.qlive)
        else:
            self.ib.cancelRealTimeBars(self.qlive)

    def haslivedata(self):
        return bool(self._storedmsg or self.qlive)

    def _load(self):
        if self.contract is None or self._state == self._ST_OVER:
            return False  # nothing can be done

        while True:
            if self._state == self._ST_LIVE:
                try:
                    msg = (self._storedmsg.pop(None, None) or
                           self.qlive.get(block=False))
                except queue.Empty:
                    if True:
                        return None
                ret = self._load_md(msg)
                if ret:
                    return True
                if msg is None:  # Conn broken during historical/backfilling
                    continue

    def _st_start(self):
        if self.p.historical:
            self.put_notification(self.DELAYED)
            dtend = None
            if self.todate < float('inf'):
                dtend = num2date(self.todate)

            dtbegin = None
            if self.fromdate > float('-inf'):
                dtbegin = num2date(self.fromdate)

            self.qhist = self.ib.reqHistoricalDataEx(
                contract=self.contract, enddate=dtend, begindate=dtbegin,
                timeframe=self._timeframe, compression=self._compression,
                what=self.p.what, useRTH=self.p.useRTH, tz=self._tz,
                sessionend=self.p.sessionend)

            self._state = self._ST_HISTORBACK
            return True  # continue before

            # Live is requested. Esto se comento todo porque no nos importa
            # if not self.ib.reconnect(resub=True):
            #   self.put_notification(self.DISCONNECTED)
            #   self._state = self._ST_OVER
            #   return False  # failed - was so

        self._statelivereconn = self.p.backfill_start
        if self.p.backfill_start:
            self.put_notification(self.DELAYED)

        self._state = self._ST_LIVE
        return True  # no return before - implicit continue

    def _load_rtbar(self, rtbar, hist=False):
        # A complete 5 second bar made of real-time ticks is delivered and
        # contains open/high/low/close/volume prices
        # The historical data has the same data but with 'date' instead of
        # 'time' for datetime
        dt = date2num(rtbar.time if not hist else rtbar.date)
        if dt < self.lines.datetime[-1] and not self.p.latethrough:
            return False  # cannot deliver earlier than already delivered

        self.lines.datetime[0] = dt
        # Put the tick into the bar
        self.lines.open[0] = rtbar.open
        self.lines.high[0] = rtbar.high
        self.lines.low[0] = rtbar.low
        self.lines.close[0] = rtbar.close
        self.lines.volume[0] = rtbar.volume
        self.lines.openinterest[0] = 0

        return True

    def _load_rtvolume(self, rtvol):
        # A single tick is delivered and is therefore used for the entire set
        # of prices. Ideally the
        # contains open/high/low/close/volume prices
        # Datetime transformation
        dt = date2num(rtvol.datetime)
        if dt < self.lines.datetime[-1] and not self.p.latethrough:
            return False  # cannot deliver earlier than already delivered

        self.lines.datetime[0] = dt

        # Put the tick into the bar
        tick = rtvol.price
        self.lines.open[0] = tick
        self.lines.high[0] = tick
        self.lines.low[0] = tick
        self.lines.close[0] = tick
        self.lines.volume[0] = rtvol.size
        self.lines.openinterest[0] = 0

        return True

    @abc.abstractmethod
    def _load_md(self, md):
        pass

    def get_tick(self):
        px = self.lines.last_px[0] if not math.isnan(self.lines.last_px[0]) and self.lines.last_px[0] != 0 else \
            self.lines.close_px[0]
        ticks = None
        if self.instrument_type == 'Titulos':
            ticks = self.variacion_rf
        else:
            ticks = self.variacion_rv

        if px is not None and not math.isnan(px):
            for key in ticks.keys():
                if px <= key:
                    return ticks[key]
        return ticks[list(ticks.keys())[-1]]



import traceback

from .event.event_factory import eventFactory
try:
    from arquants.feed.aggregated_data import *
    from arquants.feed.disaggregated_data import *
    from arquants.feed.top_ok_book_data import *
except ImportError as e:
    print(traceback.print_exc())
    print("Error importing data")

try:
    from arquants.brokers.aq_broker import *
except ImportError:
    print("Error importing broker")

from .strategy import *


"""
     ___   _____    _____     _   _       ___   __   _   _____   _____
    /   | |  _  \  /  _  \   | | | |     /   | |  \ | | |_   _| /  ___/
   / /| | | |_| |  | | | |   | | | |    / /| | |   \| |   | |   | |___
  / / | | |  _  /  | | | |   | | | |   / / | | | |\   |   | |   \___  \
 / /  | | | | \ \  | |_| |_  | |_| |  / /  | | | | \  |   | |    ___| |
/_/   |_| |_|  \_\ \_______| \_____/ /_/   |_| |_|  \_|   |_|   /_____/

"""

__title__ = 'ArQuants Algorithm Engine'
__author__ = 'arquants'
__license__ = 'BSD 2-Clause'








import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
import os
from aq_lib.configuration.configuration import Configuration
from aq_lib.manager.public_handler import PublicAQHandlerLayer
from aq_lib.manager.wsclient import manager_queue

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
CONFIG_PATH = ROOT_DIR + '/config.ini'
configuration = Configuration(CONFIG_PATH)

pool = ThreadPoolExecutor()
sheet_queue = Queue()

##################################################################
# Setting
ws_endpoint_md = os.environ.get('WS_OR_URL', 'wss://api.arquants.trading/live/market_data')
ws_endpoint_er = os.environ.get('WS_ER_URL', 'wss://api.arquants.trading/live/or')
api_endpoint = os.environ.get('API_URL', 'https://api.arquants.trading/api/v1/')

aq_username = os.environ.get('USERNAME', None)
aq_password = os.environ.get('PASSWORD', None)

if aq_password is None or aq_password is None:
    raise Exception('The environment variable USERNAME or PASSWORD was not found')


# TODO: hay que definir un único __init__ para ambos manejadores public/private (polimorfismo).
aqhandler = PublicAQHandlerLayer(aq_username, aq_password, api_endpoint, ws_endpoint_er, ws_endpoint_md, manager_queue)


def force_async(fn):
    """
        decorator:
        turns a sync function to async function using threads
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        future = pool.submit(fn, *args, **kwargs)
        return asyncio.wrap_future(future)

    return wrapper



import traceback
from multiprocessing import Manager

from apscheduler.schedulers.background import BackgroundScheduler
import threading


class Accumulator:

    def __init__(self, processor):
        manager = Manager()
        self._scheduler = BackgroundScheduler()
        self._interval = 4
        self._items = manager.list()
        self._lock = threading.Lock()
        self._processor = processor

    def process_item(self, item):
        with self._lock:
            print('agregando item a acumulator')
            self._items.append(item)

    def start(self):
        self._scheduler.start()
        self._scheduler.add_job(func=self._purge, trigger='interval', seconds=self._interval, args=[])

    def stop(self):
        self._scheduler.shutdown()

    def _purge(self):
        with self._lock:
            if len(self._items) == 0:
                return
            print('enviando item a batch')
            extracted_items = self._items[:]
            try:
                self._processor(extracted_items)
            except:
                print(traceback.format_exc())
            self._items[:] = []

# ==============================================================================

# print('starting test..')
#
# def printer_processor(message):
#     print(message)
#
# acc = Accumulator(printer_processor)
# acc.start()
#
# for i in range(20):
#     acc.process_item('foo' + str(i))
#     if not i % 5:
#         time.sleep(4)
#
# time.sleep(10)
# acc.stop()
#
# print('..done')



from multiprocessing import Process


class QueueReader(Process):
    def __init__(self, queue=None, processor=None):
        super(QueueReader, self).__init__()
        self.logger_queue = queue
        self.processor = processor

    def run(self):
        while True:
            item = self.logger_queue.get()
            print("se leyo un item en queue reader")
            self.processor.process_item(item)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import datetime

from aq_lib.model.md_entry_type import MDEntryType
from backtrader import date2num

from arquants.feed.base_data import BaseData


class TopOfBookData(BaseData):

    data_type = 'top'

    lines = ('bid_px', 'bid_qty', 'offer_px', 'offer_qty', 'last_px', 'last_qty', 'close_px', 'nominal_volume',
             'effective_volume', 'dummy_line')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.bidPxDepth = []
        self.bidQtyDepth = []
        self.offerPxDepth = []
        self.offerQtyDepth = []
        self.init_lines_depth()

    def init_lines_depth(self):
        self.bidPxDepth.append(self.lines.bid_px)
        self.bidQtyDepth.append(self.lines.bid_qty)
        self.offerPxDepth.append(self.lines.offer_px)
        self.offerQtyDepth.append(self.lines.offer_qty)

    def _load_md(self, md):
        dt = date2num(datetime.datetime.now())
        self.lines.datetime[0] = dt
        loaded = False
        for entry in md.entries:
            position = 0
            if entry.position is not None:
                position = entry.position - 1
            if entry.entry_type == MDEntryType.BID.value:
                if position <= 4:
                    self.bidPxDepth[position][0] = entry.entry_px
                    self.bidQtyDepth[position][0] = entry.entry_size
                    loaded = True
            elif entry.entry_type == MDEntryType.OFFER.value:
                if position <= 4:
                    self.offerPxDepth[position][0] = entry.entry_px
                    self.offerQtyDepth[position][0] = entry.entry_size
                    loaded = True
            elif entry.entry_type == MDEntryType.TRADE.value:
                self.lines.last_px[0] = entry.entry_px
                self.lines.last_qty[0] = entry.entry_size
                self.position.update_on_md(self.tradecontract, entry.entry_px)
                loaded = True
            elif entry.entry_type == MDEntryType.CLOSE.value:
                self.lines.close_px[0] = entry.entry_px
                loaded = True
            elif entry.entry_type == MDEntryType.TRADE_VOLUME.value:
                self.lines.effective_volume[0] = entry.entry_px
                self.lines.nominal_volume[0] = entry.entry_size
                loaded = True

        if not loaded:
            self.lines.dummy_line[0] = 0
        return True



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
###############################################################################
#
# Copyright (C) 2015, 2016, 2017 Daniel Rodriguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import collections
import datetime
import json
import uuid
from collections import OrderedDict

import backtrader as bt
from aq_lib.utils.logger import logger
from backtrader.order import Order, BuyOrder, SellOrder
from backtrader.utils.py3 import string_types, integer_types

from arquants.definitions import aqhandler
from arquants.event.event_factory import eventFactory
from arquants.observers.position_calculator import PositionCalculator

__all__ = ['BackBroker', 'BrokerBack']


class BackBroker(bt.BrokerBase):
    """Broker Simulator

      The simulation supports different order types, checking a submitted order
      cash requirements against current cash, keeping track of cash and value
      for each iteration of ``cerebro`` and keeping the current position on
      different datas.

      *cash* is adjusted on each iteration for instruments like ``futures`` for
       which a price change implies in real brokers the addition/substracion of
       cash.

      Supported order types:

        - ``Market``: to be executed with the 1st tick of the next bar (namely
          the ``open`` price)

        - ``Close``: meant for intraday in which the order is executed with the
          closing price of the last bar of the session

        - ``Limit``: executes if the given limit price is seen during the
          session

        - ``Stop``: executes a ``Market`` order if the given stop price is seen

        - ``StopLimit``: sets a ``Limit`` order in motion if the given stop
          price is seen

      Because the broker is instantiated by ``Cerebro`` and there should be
      (mostly) no reason to replace the broker, the params are not controlled
      by the user for the instance.  To change this there are two options:

        1. Manually create an instance of this class with the desired params
           and use ``cerebro.broker = instance`` to set the instance as the
           broker for the ``run`` execution

        2. Use the ``set_xxx`` to set the value using
           ``cerebro.broker.set_xxx`` where ```xxx`` stands for the name of the
           parameter to set

        .. note::

           ``cerebro.broker`` is a *property* supported by the ``getbroker``
           and ``setbroker`` methods of ``Cerebro``

      Params:

        - ``cash`` (default: ``10000``): starting cash

        - ``commission`` (default: ``CommInfoBase(percabs=True)``)
          base commission scheme which applies to all assets

        - ``checksubmit`` (default: ``True``)
          check margin/cash before accepting an order into the system

        - ``eosbar`` (default: ``False``):
          With intraday bars consider a bar with the same ``time`` as the end
          of session to be the end of the session. This is not usually the
          case, because some bars (final auction) are produced by many
          exchanges for many products for a couple of minutes after the end of
          the session

        - ``eosbar`` (default: ``False``):
          With intraday bars consider a bar with the same ``time`` as the end
          of session to be the end of the session. This is not usually the
          case, because some bars (final auction) are produced by many
          exchanges for many products for a couple of minutes after the end of
          the session

        - ``filler`` (default: ``None``)

          A callable with signature: ``callable(order, price, ago)``

            - ``order``: obviously the order in execution. This provides access
              to the *data* (and with it the *ohlc* and *volume* values), the
              *execution type*, remaining size (``order.executed.remsize``) and
              others.

              Please check the ``Order`` documentation and reference for things
              available inside an ``Order`` instance

            - ``price`` the price at which the order is going to be executed in
              the ``ago`` bar

            - ``ago``: index meant to be used with ``order.data`` for the
              extraction of the *ohlc* and *volume* prices. In most cases this
              will be ``0`` but on a corner case for ``Close`` orders, this
              will be ``-1``.

              In order to get the bar volume (for example) do: ``volume =
              order.data.voluume[ago]``

          The callable must return the *executed size* (a value >= 0)

          The callable may of course be an object with ``__call__`` matching
          the aforementioned signature

          With the default ``None`` orders will be completely executed in a
          single shot

        - ``slip_perc`` (default: ``0.0``) Percentage in absolute termns (and
          positive) that should be used to slip prices up/down for buy/sell
          orders

          Note:

            - ``0.01`` is ``1%``

            - ``0.001`` is ``0.1%``

        - ``slip_fixed`` (default: ``0.0``) Percentage in units (and positive)
          that should be used to slip prices up/down for buy/sell orders

          Note: if ``slip_perc`` is non zero, it takes precendence over this.

        - ``slip_open`` (default: ``False``) whether to slip prices for order
          execution which would specifically used the *opening* price of the
          next bar. An example would be ``Market`` order which is executed with
          the next available tick, i.e: the opening price of the bar.

          This also applies to some of the other executions, because the logic
          tries to detect if the *opening* price would match the requested
          price/execution type when moving to a new bar.

        - ``slip_match`` (default: ``True``)

          If ``True`` the broker will offer a match by capping slippage at
          ``high/low`` prices in case they would be exceeded.

          If ``False`` the broker will not match the order with the current
          prices and will try execution during the next iteration

        - ``slip_limit`` (default: ``True``)

          ``Limit`` orders, given the exact match price requested, will be
          matched even if ``slip_match`` is ``False``.

          This option controls that behavior.

          If ``True``, then ``Limit`` orders will be matched by capping prices
          to the ``limit`` / ``high/low`` prices

          If ``False`` and slippage exceeds the cap, then there will be no
          match

        - ``slip_out`` (default: ``False``)

          Provide *slippage* even if the price falls outside the ``high`` -
          ``low`` range.

        - ``coc`` (default: ``False``)

          *Cheat-On-Close* Setting this to ``True`` with ``set_coc`` enables
           matching a ``Market`` order to the closing price of the bar in which
           the order was issued. This is actually *cheating*, because the bar
           is *closed* and any order should first be matched against the prices
           in the next bar

        - ``coo`` (default: ``False``)

          *Cheat-On-Open* Setting this to ``True`` with ``set_coo`` enables
           matching a ``Market`` order to the opening price, by for example
           using a timer with ``cheat`` set to ``True``, because such a timer
           gets executed before the broker has evaluated

        - ``int2pnl`` (default: ``True``)

          Assign generated interest (if any) to the profit and loss of
          operation that reduces a position (be it long or short). There may be
          cases in which this is undesired, because different strategies are
          competing and the interest would be assigned on a non-deterministic
          basis to any of them.

        - ``shortcash`` (default: ``True``)

          If True then cash will be increased when a stocklike asset is shorted
          and the calculated value for the asset will be negative.

          If ``False`` then the cash will be deducted as operation cost and the
          calculated value will be positive to end up with the same amount

        - ``fundstartval`` (default: ``100.0``)

          This parameter controls the start value for measuring the performance
          in a fund-like way, i.e.: cash can be added and deducted increasing
          the amount of shares. Performance is not measured using the net
          asset value of the porftoflio but using the value of the fund

        - ``fundmode`` (default: ``False``)

          If this is set to ``True`` analyzers like ``TimeReturn`` can
          automatically calculate returns based on the fund value and not on
          the total net asset value

    """
    params = (
        ('cash', 10000.0),
        ('checksubmit', True),
        ('eosbar', False),
        ('filler', None),
        # slippage options
        ('slip_perc', 0.0),
        ('slip_fixed', 0.0),
        ('slip_open', False),
        ('slip_match', True),
        ('slip_limit', True),
        ('slip_out', False),
        ('coc', False),
        ('coo', False),
        ('int2pnl', True),
        ('shortcash', True),
        ('fundstartval', 100.0),
        ('fundmode', False),
    )

    def __init__(self):
        super(BackBroker, self).__init__()
        self._userhist = []
        self._fundhist = []
        # share_value, net asset value
        self._fhistlast = [float('NaN'), float('NaN')]

    def init(self):
        super(BackBroker, self).init()
        self.startingcash = self.cash = self.p.cash
        self._value = self.cash
        self._valuemkt = 0.0  # no open position

        self._valuelever = 0.0  # no open position
        self._valuemktlever = 0.0  # no open position

        self._leverage = 1.0  # initially nothing is open
        self._unrealized = 0.0  # no open position

        self.orders = list()  # will only be appending
        self.pending = collections.deque()  # popleft and append(right)
        self._toactivate = collections.deque()  # to activate in next cycle

        self.positions = collections.defaultdict(PositionCalculator)
        self.d_credit = collections.defaultdict(float)  # credit per data
        self.notifs = collections.deque()

        self.submitted = collections.deque()

        # to keep dependent orders if needed
        self._pchildren = collections.defaultdict(collections.deque)

        self._ocos = dict()
        self._ocol = collections.defaultdict(list)

        self._fundval = self.p.fundstartval
        self._fundshares = self.p.cash / self._fundval
        self._cash_addition = collections.deque()
        self.orderUpdatedEvents = OrderedDict()

    def get_notification(self):
        try:
            return self.notifs.popleft()
        except IndexError:
            pass

        return None

    def set_fundmode(self, fundmode, fundstartval=None):
        """Set the actual fundmode (True or False)

        If the argument fundstartval is not ``None``, it will used
        """
        self.p.fundmode = fundmode
        if fundstartval is not None:
            self.set_fundstartval(fundstartval)

    def get_fundmode(self):
        """Returns the actual fundmode (True or False)"""
        return self.p.fundmode

    fundmode = property(get_fundmode, set_fundmode)

    def set_fundstartval(self, fundstartval):
        """Set the starting value of the fund-like performance tracker"""
        self.p.fundstartval = fundstartval

    def set_int2pnl(self, int2pnl):
        """Configure assignment of interest to profit and loss"""
        self.p.int2pnl = int2pnl

    def set_coc(self, coc):
        """Configure the Cheat-On-Close method to buy the close on order bar"""
        self.p.coc = coc

    def set_coo(self, coo):
        """Configure the Cheat-On-Open method to buy the close on order bar"""
        self.p.coo = coo

    def set_shortcash(self, shortcash):
        """Configure the shortcash parameters"""
        self.p.shortcash = shortcash

    def set_slippage_perc(self, perc,
                          slip_open=True, slip_limit=True,
                          slip_match=True, slip_out=False):
        """Configure slippage to be percentage based"""
        self.p.slip_perc = perc
        self.p.slip_fixed = 0.0
        self.p.slip_open = slip_open
        self.p.slip_limit = slip_limit
        self.p.slip_match = slip_match
        self.p.slip_out = slip_out

    def set_slippage_fixed(self, fixed,
                           slip_open=True, slip_limit=True,
                           slip_match=True, slip_out=False):
        """Configure slippage to be fixed points based"""
        self.p.slip_perc = 0.0
        self.p.slip_fixed = fixed
        self.p.slip_open = slip_open
        self.p.slip_limit = slip_limit
        self.p.slip_match = slip_match
        self.p.slip_out = slip_out

    def set_filler(self, filler):
        """Sets a volume filler for volume filling execution"""
        self.p.filler = filler

    def set_checksubmit(self, checksubmit):
        """Sets the checksubmit parameter"""
        self.p.checksubmit = checksubmit

    def set_eosbar(self, eosbar):
        """Sets the eosbar parameter (alias: ``seteosbar``"""
        self.p.eosbar = eosbar

    seteosbar = set_eosbar

    def get_cash(self):
        """Returns the current cash (alias: ``getcash``)"""
        return self.cash

    getcash = get_cash

    def set_cash(self, cash):
        """Sets the cash parameter (alias: ``setcash``)"""
        self.startingcash = self.cash = self.p.cash = cash
        self._value = cash

    setcash = set_cash

    def add_cash(self, cash):
        """Add/Remove cash to the system (use a negative value to remove)"""
        self._cash_addition.append(cash)

    def get_fundshares(self):
        """Returns the current number of shares in the fund-like mode"""
        return self._fundshares

    fundshares = property(get_fundshares)

    def get_fundvalue(self):
        """Returns the Fund-like share value"""
        return self._fundval

    fundvalue = property(get_fundvalue)

    def cancel(self, order, bracket=False):
        try:
            self.pending.remove(order)
        except ValueError:
            # If the list didn't have the element we didn't cancel anything
            return False

        order.cancel()
        self.notify(order)
        self._ococheck(order)
        if not bracket:
            self._bracketize(order, cancel=True)
        return True

    def get_value(self, datas=None, mkt=False, lever=False):
        """Returns the portfolio value of the given datas (if datas is ``None``, then
        the total portfolio value will be returned (alias: ``getvalue``)
        """
        if datas is None:
            if mkt:
                return self._valuemkt if not lever else self._valuemktlever

            return self._value if not lever else self._valuelever

        return self._get_value(datas=datas, lever=lever)

    getvalue = get_value

    def get_value_lever(self, datas=None, mkt=False):
        return self.get_value(datas=datas, mkt=mkt)

    def _get_value(self, datas=None, lever=False):
        pos_value = 0.0
        pos_value_unlever = 0.0
        unrealized = 0.0

        while self._cash_addition:
            c = self._cash_addition.popleft()
            self._fundshares += c / self._fundval
            self.cash += c

        for data in datas or self.positions:
            comminfo = self.getcommissioninfo(data)
            position = self.positions[data]
            # use valuesize:  returns raw value, rather than negative adj val
            if not self.p.shortcash:
                dvalue = comminfo.getvalue(position, data.close[0])
            else:
                dvalue = comminfo.getvaluesize(position.size, data.close[0])

            dunrealized = comminfo.profitandloss(position.size, position.price,
                                                 data.close[0])
            if datas and len(datas) == 1:
                if lever and dvalue > 0:
                    dvalue -= dunrealized
                    return (dvalue / comminfo.get_leverage()) + dunrealized
                return dvalue  # raw data value requested, short selling is neg

            if not self.p.shortcash:
                dvalue = abs(dvalue)  # short selling adds value in this case

            pos_value += dvalue
            unrealized += dunrealized

            if dvalue > 0:  # long position - unlever
                dvalue -= dunrealized
                pos_value_unlever += (dvalue / comminfo.get_leverage())
                pos_value_unlever += dunrealized
            else:
                pos_value_unlever += dvalue

        if not self._fundhist:
            self._value = v = self.cash + pos_value_unlever
            self._fundval = self._value / self._fundshares  # update fundvalue
        else:
            # Try to fetch a value
            fval, fvalue = self._process_fund_history()

            self._value = fvalue
            self.cash = fvalue - pos_value_unlever
            self._fundval = fval
            self._fundshares = fvalue / fval
            lev = pos_value / (pos_value_unlever or 1.0)

            # update the calculated values above to the historical values
            pos_value_unlever = fvalue
            pos_value = fvalue * lev

        self._valuemkt = pos_value_unlever

        self._valuelever = self.cash + pos_value
        self._valuemktlever = pos_value

        self._leverage = pos_value / (pos_value_unlever or 1.0)
        self._unrealized = unrealized

        return self._value if not lever else self._valuelever

    def get_leverage(self):
        return self._leverage

    def get_orders_open(self, safe=False):
        """Returns an iterable with the orders which are still open (either not
        executed or partially executed

        The orders returned must not be touched.

        If order manipulation is needed, set the parameter ``safe`` to True
        """
        if safe:
            os = [x.clone() for x in self.pending]
        else:
            os = [x for x in self.pending]

        return os

    def getposition(self, data, clone=False):
        """Returns the current position status (a ``Position`` instance) for
        the given ``data``"""
        return self.positions[data]

    def orderstatus(self, order):
        try:
            o = self.orders.index(order)
        except ValueError:
            o = order

        return o.status

    def _take_children(self, order):
        oref = order.ref
        pref = getattr(order.parent, 'ref', oref)  # parent ref or self

        if oref != pref:
            if pref not in self._pchildren:
                order.reject()  # parent not there - may have been rejected
                self.notify(order)  # reject child, notify
                return None

        return pref

    def submit(self, order, check=True):
        pref = self._take_children(order)
        if pref is None:  # order has not been taken
            return order

        pc = self._pchildren[pref]
        pc.append(order)  # store in parent/children queue

        if order.transmit:  # if single order, sent and queue cleared
            # if parent-child, the parent will be sent, the other kept
            rets = [self.transmit(x, check=check) for x in pc]
            return rets[-1]  # last one is the one triggering transmission

        return order

    def transmit(self, order, check=True):
        if check and self.p.checksubmit:
            order.submit()
            self.submitted.append(order)
            self.orders.append(order)
            self.notify(order)
        else:
            self.submit_accept(order)

        return order

    def check_submitted(self):
        cash = self.cash
        positions = dict()

        while self.submitted:
            order = self.submitted.popleft()

            if self._take_children(order) is None:  # children not taken
                continue

            self.getcommissioninfo(order.data)

            position = positions.setdefault(order.data, self.positions[order.data].clone())

            # pseudo-execute the order to get the remaining cash after exec
            cash = self._execute(order, cash=cash, position=position)

            if cash >= 0.0:
                self.submit_accept(order)
                continue

            order.margin()
            self.notify(order)
            self._ococheck(order)
            self._bracketize(order, cancel=True)

    def submit_accept(self, order):
        order.pannotated = None
        order.submit()
        order.accept()
        self.pending.append(order)
        self.notify(order)

    def _bracketize(self, order, cancel=False):
        oref = order.ref
        pref = getattr(order.parent, 'ref', oref)
        parent = oref == pref

        pc = self._pchildren[pref]  # defdict - guaranteed
        if cancel or not parent:  # cancel left or child exec -> cancel other
            while pc:
                self.cancel(pc.popleft(), bracket=True)  # idempotent

            del self._pchildren[pref]  # defdict guaranteed

        else:  # not cancel -> parent exec'd
            pc.popleft()  # remove parent
            for o in pc:  # activate childnre
                self._toactivate.append(o)

    def _ococheck(self, order):
        # ocoref = self._ocos[order.ref] or order.ref  # a parent or self
        parentref = self._ocos[order.ref]
        ocoref = self._ocos.get(parentref, None)
        ocol = self._ocol.pop(ocoref, None)
        if ocol:
            for i in range(len(self.pending) - 1, -1, -1):
                o = self.pending[i]
                if o is not None and o.ref in ocol:
                    del self.pending[i]
                    o.cancel()
                    self.notify(o)

    def _ocoize(self, order, oco):
        oref = order.ref
        if oco is None:
            self._ocos[oref] = oref  # current order is parent
            self._ocol[oref].append(oref)  # create ocogroup
        else:
            ocoref = self._ocos[oco.ref]  # ref to group leader
            self._ocos[oref] = ocoref  # ref to group leader
            self._ocol[ocoref].append(oref)  # add to group

    def add_order_history(self, orders, notify=True):
        oiter = iter(orders)
        o = next(oiter, None)
        self._userhist.append([o, oiter, notify])

    def set_fund_history(self, fund):
        # iterable with the following pro item
        # [datetime, share_value, net asset value]
        fiter = iter(fund)
        f = list(next(fiter))  # must not be empty
        self._fundhist = [f, fiter]
        # self._fhistlast = f[1:]

        self.set_cash(float(f[2]))

    def buy(self, owner, data,
            size, price=None, plimit=None,
            exectype=None, valid=None, tradeid=0, oco=None,
            trailamount=None, trailpercent=None,
            parent=None, transmit=True,
            histnotify=False, _checksubmit=True,
            **kwargs):

        order = BuyOrder(owner=owner, data=data, size=size, price=price, pricelimit=plimit, exectype=exectype,
                         valid=valid, tradeid=tradeid, trailamount=trailamount, trailpercent=trailpercent,
                         parent=parent, transmit=transmit, histnotify=histnotify)
        order.m_orderId = self.generate_id()
        order.addinfo(**kwargs)
        self._ocoize(order, oco)

        return self.submit(order, check=_checksubmit)

    def sell(self, owner, data,
             size, price=None, plimit=None,
             exectype=None, valid=None, tradeid=0, oco=None,
             trailamount=None, trailpercent=None,
             parent=None, transmit=True,
             histnotify=False, _checksubmit=True,
             **kwargs):

        order = SellOrder(owner=owner, data=data, size=size, price=price, pricelimit=plimit, exectype=exectype,
                          valid=valid, tradeid=tradeid, trailamount=trailamount, trailpercent=trailpercent,
                          parent=parent, transmit=transmit, histnotify=histnotify)
        order.m_orderId = self.generate_id()
        order.addinfo(**kwargs)
        self._ocoize(order, oco)

        return self.submit(order, check=_checksubmit)

    # TODO: Aca hay que gomosear con la posicion
    def _execute(self, order, ago=None, price=None, cash=None, position=None,
                 dtcoc=None):
        # ago = None is used a flag for pseudo execution
        if ago is not None and price is None:
            return  # no psuedo exec no price - no execution

        if self.p.filler is None or ago is None:
            # Order gets full size or pseudo-execution
            size = order.executed.remsize
        else:
            # Execution depends on volume filler
            size = self.p.filler(order, price, ago)
            if not order.isbuy():
                size = -size

        # Get comminfo object for the data
        comminfo = self.getcommissioninfo(order.data)

        # Check if something has to be compensated
        if order.data._compensate is not None:
            data = order.data._compensate
            cinfocomp = self.getcommissioninfo(data)  # for actual commission
        else:
            data = order.data
            cinfocomp = comminfo

        # Adjust position with operation size
        if ago is not None:
            # Real execution with date
            position = self.positions[data]
            pprice_orig = position.price

            psize, pprice, opened, closed = position.pseudoupdate(size, price)

            # if part/all of a position has been closed, then there has been
            # a profitandloss ... record it
            pnl = comminfo.profitandloss(-closed, pprice_orig, price)
            cash = self.cash
        else:
            pnl = 0
            price = pprice_orig = order.created.price
            psize, pprice, opened, closed = position.update(size, price)  # TODO Aca hace update position

        # "Closing" totally or partially is possible. Cash may be re-injected
        if closed:
            # Adjust to returned value for closed items & acquired opened items
            if self.p.shortcash:
                closedvalue = comminfo.getvaluesize(-closed, pprice_orig)
            else:
                closedvalue = comminfo.getoperationcost(closed, pprice_orig)

            closecash = closedvalue
            if closedvalue > 0:  # long position closed
                closecash /= comminfo.get_leverage()  # inc cash with lever

            cash += closecash + pnl * comminfo.stocklike
            # Calculate and substract commission
            closedcomm = comminfo.getcommission(closed, price)
            cash -= closedcomm

            if ago is not None:
                # Cashadjust closed contracts: prev close vs exec price
                # The operation can inject or take cash out
                cash += comminfo.cashadjust(-closed,
                                            position.adjbase,
                                            price)

                # Update system cash
                self.cash = cash
        else:
            closedvalue = closedcomm = 0.0

        popened = opened
        if opened:
            if self.p.shortcash:
                openedvalue = comminfo.getvaluesize(opened, price)
            else:
                openedvalue = comminfo.getoperationcost(opened, price)

            opencash = openedvalue
            if openedvalue > 0:  # long position being opened
                opencash /= comminfo.get_leverage()  # dec cash with level

            cash -= opencash  # original behavior

            openedcomm = cinfocomp.getcommission(opened, price)
            cash -= openedcomm

            if cash < 0.0:
                # execution is not possible - nullify
                opened = 0
                openedvalue = openedcomm = 0.0

            elif ago is not None:  # real execution
                if abs(psize) > abs(opened):
                    # some futures were opened - adjust the cash of the
                    # previously existing futures to the operation price and
                    # use that as new adjustment base, because it already is
                    # for the new futures At the end of the cycle the
                    # adjustment to the close price will be done for all open
                    # futures from a common base price with regards to the
                    # close price
                    adjsize = psize - opened
                    cash += comminfo.cashadjust(adjsize,
                                                position.adjbase, price)

                # record adjust price base for end of bar cash adjustment
                position.adjbase = price

                # update system cash - checking if opened is still != 0
                self.cash = cash
        else:
            openedvalue = openedcomm = 0.0

        if ago is None:
            # return cash from pseudo-execution
            return cash

        execsize = closed + opened

        if execsize:
            # Confimrm the operation to the comminfo object
            comminfo.confirmexec(execsize, price)

            # do a real position update if something was executed
            position.update(execsize, price, data.datetime.datetime())
            if order.data.last_px is not None:
                position.updateLastPrice(order.data.last_px)
            if closed and self.p.int2pnl:  # Assign accumulated interest data
                closedcomm += self.d_credit.pop(data, 0.0)

            # Execute and notify the order
            order.execute(dtcoc or data.datetime[ago],
                          execsize, price,
                          closed, closedvalue, closedcomm,
                          opened, openedvalue, openedcomm,
                          comminfo.margin, pnl,
                          psize, pprice)

            order.addcomminfo(comminfo)

            if price != order.price:
                order.execprice = price
            self.notify(order)
            self._ococheck(order)
        if popened and not opened:
            # opened was not executed - not enough cash
            order.margin()
            self.notify(order)
            self._ococheck(order)
            self._bracketize(order, cancel=True)

    def notify(self, order):
        self.notifs.append(order.clone())
        event = eventFactory.createOrderUpdatedEventFromOrder(order)

        if self.passive_trading:
            aqhandler.send(json.dumps(event.to_dict()), self.delivery_routing_key)
        else:
            self.orderUpdatedEvents[event.clientOrderID] = event

    def stop(self):
        if not self.passive_trading:
            aqhandler.send(json.dumps(eventFactory.createBatchOrderUpdatedEvent(self.orderUpdatedEvents).to_dict()),
                           self.delivery_routing_key)
            positions = dict()
            for data in self.cerebro.datas:
                positions[data.tradecontract] = self.getposition(data, True).get_position()
            aqhandler.send(json.dumps(eventFactory.createPositionUpdatedEvent(positions).to_dict()),
                           self.delivery_routing_key)

    def _try_exec_historical(self, order):
        self._execute(order, ago=0, price=order.created.price)

    def send_bulk_new_orders(self, orders):
        pass

    def _try_exec_market(self, order, popen, phigh, plow):
        0
        if self.p.coc and order.info.get('coc', True):
            dtcoc = order.created.dt
            exprice = order.created.pclose
        else:
            if not self.p.coo and order.data.datetime[0] <= order.created.dt:
                return    # can only execute after creation time

            dtcoc = None
            exprice = popen

        if order.isbuy():
            p = self._slip_up(phigh, exprice, doslip=self.p.slip_open)
        else:
            p = self._slip_down(plow, exprice, doslip=self.p.slip_open)

        self._execute(order, ago=0, price=p, dtcoc=dtcoc)

    def _try_exec_close(self, order, pclose):
        # pannotated allows to keep track of the closing bar if there is no
        # information which lets us know that the current bar is the closing
        # bar (like matching end of session bar)
        # The actual matching will be done one bar afterwards but using the
        # information from the actual closing bar

        dt0 = order.data.datetime[0]
        # don't use "len" -> in replay the close can be reached with same len
        if dt0 > order.created.dt:  # can only execute after creation time
            # or (self.p.eosbar and dt0 == order.dteos):
            if dt0 >= order.dteos:
                # past the end of session or right at it and eosbar is True
                if order.pannotated and dt0 > order.dteos:
                    ago = -1
                    execprice = order.pannotated
                else:
                    ago = 0
                    execprice = pclose

                self._execute(order, ago=ago, price=execprice)
                return

        # If no exexcution has taken place ... annotate the closing price
        order.pannotated = pclose

    def _try_exec_limit(self, order, popen, phigh, plow, plimit):
        if order.isbuy():
            if plimit >= popen:
                # open smaller/equal than requested - buy cheaper
                pmax = min(phigh, plimit)
                p = self._slip_up(pmax, popen, doslip=self.p.slip_open,
                                  lim=True)
                self._execute(order, ago=0, price=p)
            elif plimit >= plow:
                # day low below req price ... match limit price
                self._execute(order, ago=0, price=plimit)

        else:  # Sell
            if plimit <= popen:
                # open greater/equal than requested - sell more expensive
                max(plow, plimit)
                p = self._slip_down(plimit, popen, doslip=self.p.slip_open,
                                    lim=True)
                self._execute(order, ago=0, price=p)
            elif plimit <= phigh:
                # day high above req price ... match limit price
                self._execute(order, ago=0, price=plimit)

    def _try_exec_stop(self, order, popen, phigh, plow, pcreated, pclose):
        if order.isbuy():
            if popen >= pcreated:
                # price penetrated with an open gap - use open
                p = self._slip_up(phigh, popen, doslip=self.p.slip_open)
                self._execute(order, ago=0, price=p)
            elif phigh >= pcreated:
                # price penetrated during the session - use trigger price
                p = self._slip_up(phigh, pcreated)
                self._execute(order, ago=0, price=p)

        else:  # Sell
            if popen <= pcreated:
                # price penetrated with an open gap - use open
                p = self._slip_down(plow, popen, doslip=self.p.slip_open)
                self._execute(order, ago=0, price=p)
            elif plow <= pcreated:
                # price penetrated during the session - use trigger price
                p = self._slip_down(plow, pcreated)
                self._execute(order, ago=0, price=p)

        # not (completely) executed and trailing stop
        if order.alive() and order.exectype == Order.StopTrail:
            order.trailadjust(pclose)

    def _try_exec_stoplimit(self, order,
                            popen, phigh, plow, pclose,
                            pcreated, plimit):
        if order.isbuy():
            if popen >= pcreated:
                order.triggered = True
                self._try_exec_limit(order, popen, phigh, plow, plimit)

            elif phigh >= pcreated:
                # price penetrated upwards during the session
                order.triggered = True
                # can calculate execution for a few cases - datetime is fixed
                if popen > pclose:
                    if plimit >= pcreated:  # limit above stop trigger
                        p = self._slip_up(phigh, pcreated, lim=True)
                        self._execute(order, ago=0, price=p)
                    elif plimit >= pclose:
                        self._execute(order, ago=0, price=plimit)
                else:  # popen < pclose
                    if plimit >= pcreated:
                        p = self._slip_up(phigh, pcreated, lim=True)
                        self._execute(order, ago=0, price=p)
        else:  # Sell
            if popen <= pcreated:
                # price penetrated downwards with an open gap
                order.triggered = True
                self._try_exec_limit(order, popen, phigh, plow, plimit)

            elif plow <= pcreated:
                # price penetrated downwards during the session
                order.triggered = True
                # can calculate execution for a few cases - datetime is fixed
                if popen <= pclose:
                    if plimit <= pcreated:
                        p = self._slip_down(plow, pcreated, lim=True)
                        self._execute(order, ago=0, price=p)
                    elif plimit <= pclose:
                        self._execute(order, ago=0, price=plimit)
                else:
                    # popen > pclose
                    if plimit <= pcreated:
                        p = self._slip_down(plow, pcreated, lim=True)
                        self._execute(order, ago=0, price=p)

        # not (completely) executed and trailing stop
        if order.alive() and order.exectype == Order.StopTrailLimit:
            order.trailadjust(pclose)

    def _slip_up(self, pmax, price, doslip=True, lim=False):
        if not doslip:
            return price

        slip_perc = self.p.slip_perc
        slip_fixed = self.p.slip_fixed
        if slip_perc:
            pslip = price * (1 + slip_perc)
        elif slip_fixed:
            pslip = price + slip_fixed
        else:
            return price

        if pslip <= pmax:  # slipping can return price
            return pslip
        elif self.p.slip_match or (lim and self.p.slip_limit):
            if not self.p.slip_out:
                return pmax

            return pslip  # non existent price

        return None  # no price can be returned

    def _slip_down(self, pmin, price, doslip=True, lim=False):
        if not doslip:
            return price

        slip_perc = self.p.slip_perc
        slip_fixed = self.p.slip_fixed
        if slip_perc:
            pslip = price * (1 - slip_perc)
        elif slip_fixed:
            pslip = price - slip_fixed
        else:
            return price

        if pslip >= pmin:  # slipping can return price
            return pslip
        elif self.p.slip_match or (lim and self.p.slip_limit):
            if not self.p.slip_out:
                return pmin

            return pslip  # non existent price

        return None  # no price can be returned

    def _try_exec(self, order):
        data = order.data

        popen = getattr(data, 'tick_open', None)
        if popen is None:
            popen = data.open[0]
        phigh = getattr(data, 'tick_high', None)
        if phigh is None:
            phigh = data.high[0]
        plow = getattr(data, 'tick_low', None)
        if plow is None:
            plow = data.low[0]
        pclose = getattr(data, 'tick_close', None)
        if pclose is None:
            pclose = data.close[0]

        pcreated = order.created.price
        plimit = order.created.pricelimit

        if order.exectype == Order.Market:
            self._try_exec_market(order, popen, phigh, plow)

        elif order.exectype == Order.Close:
            self._try_exec_close(order, pclose)

        elif order.exectype == Order.Limit:
            self._try_exec_limit(order, popen, phigh, plow, pcreated)

        elif (order.triggered and
              order.exectype in [Order.StopLimit, Order.StopTrailLimit]):
            self._try_exec_limit(order, popen, phigh, plow, plimit)

        elif order.exectype in [Order.Stop, Order.StopTrail]:
            self._try_exec_stop(order, popen, phigh, plow, pcreated, pclose)

        elif order.exectype in [Order.StopLimit, Order.StopTrailLimit]:
            self._try_exec_stoplimit(order,
                                     popen, phigh, plow, pclose,
                                     pcreated, plimit)

        elif order.exectype == Order.Historical:
            self._try_exec_historical(order)

    def _process_fund_history(self):
        fhist = self._fundhist  # [last element, iterator]
        f, funds = fhist
        if not f:
            return self._fhistlast

        dt = f[0]  # date/datetime instance
        if isinstance(dt, string_types):
            dtfmt = '%Y-%m-%d'
            if 'T' in dt:
                dtfmt += 'T%H:%M:%S'
                if '.' in dt:
                    dtfmt += '.%f'
            dt = datetime.datetime.strptime(dt, dtfmt)
            f[0] = dt  # update value

        elif isinstance(dt, datetime.datetime):
            pass
        elif isinstance(dt, datetime.date):
            dt = datetime.datetime(year=dt.year, month=dt.month, day=dt.day)
            f[0] = dt  # Update the value

        # Synchronization with the strategy is not possible because the broker
        # is called before the strategy advances. The 2 lines below would do it
        # if possible
        # st0 = self.cerebro.runningstrats[0]
        # if dt <= st0.datetime.datetime():
        if dt <= self.cerebro._dtmaster:
            self._fhistlast = f[1:]
            fhist[0] = list(next(funds, []))

        return self._fhistlast

    def _process_order_history(self):
        for uhist in self._userhist:
            uhorder, uhorders, uhnotify = uhist
            while uhorder is not None:
                uhorder = list(uhorder)  # to support assignment (if tuple)
                try:
                    dataidx = uhorder[3]  # 2nd field
                except IndexError:
                    dataidx = None  # Field not present, use default

                if dataidx is None:
                    d = self.cerebro.datas[0]
                elif isinstance(dataidx, integer_types):
                    d = self.cerebro.datas[dataidx]
                else:  # assume string
                    d = self.cerebro.datasbyname[dataidx]

                if not len(d):
                    break  # may start later as oter data feeds

                dt = uhorder[0]  # date/datetime instance
                if isinstance(dt, string_types):
                    dtfmt = '%Y-%m-%d'
                    if 'T' in dt:
                        dtfmt += 'T%H:%M:%S'
                        if '.' in dt:
                            dtfmt += '.%f'
                    dt = datetime.datetime.strptime(dt, dtfmt)
                    uhorder[0] = dt
                elif isinstance(dt, datetime.datetime):
                    pass
                elif isinstance(dt, datetime.date):
                    dt = datetime.datetime(year=dt.year,
                                           month=dt.month,
                                           day=dt.day)
                    uhorder[0] = dt

                if dt > d.datetime.datetime():
                    break  # cannot execute yet 1st in queue, stop processing

                size = uhorder[1]
                price = uhorder[2]
                owner = self.cerebro.runningstrats[0]
                if size > 0:
                    self.buy(owner=owner, data=d,
                                 size=size, price=price,
                                 exectype=Order.Historical,
                                 histnotify=uhnotify,
                                 _checksubmit=False)

                elif size < 0:
                    self.sell(owner=owner, data=d,
                                  size=abs(size), price=price,
                                  exectype=Order.Historical,
                                  histnotify=uhnotify,
                                  _checksubmit=False)

                # update to next potential order
                uhist[0] = uhorder = next(uhorders, None)

    def next(self):
        while self._toactivate:
            self._toactivate.popleft().activate()

        if self.p.checksubmit:
            self.check_submitted()

        # Discount any cash for positions hold
        # credit = 0.0
        # for data, pos in self.positions.items():
        #     if pos:
        #         comminfo = self.getcommissioninfo(data)
        #         dt0 = data.datetime.datetime()
        #         dcredit = comminfo.get_credit_interest(data, pos, dt0)
        #         self.d_credit[data] += dcredit
        #         credit += dcredit
        #         pos.datetime = dt0  # mark last credit operation

        # self.cash -= credit

        self._process_order_history()

        # Iterate once over all elements of the pending queue
        self.pending.append(None)
        while True:
            order = self.pending.popleft()
            if order is None:
                break

            if order.expire():
                self.notify(order)
                self._ococheck(order)
                self._bracketize(order, cancel=True)

            elif not order.active():
                self.pending.append(order)  # cannot yet be processed

            else:
                self._try_exec(order)
                if order.alive():
                    self.pending.append(order)

                elif order.status == Order.Completed:
                    # a bracket parent order may have been executed
                    self._bracketize(order)

        # Operations have been executed ... adjust cash end of bar
        for data, pos in self.positions.items():
            # futures change cash every bar
            if pos:
                comminfo = self.getcommissioninfo(data)
                self.cash += comminfo.cashadjust(pos.size,
                                                 pos.adjbase,
                                                 data.close[0])
                # record the last adjustment price
                pos.adjbase = data.close[0]

        self._get_value()  # update value

    @staticmethod
    def generate_id():
        cl_ord_id = uuid.uuid4().hex[:16]
        logger.debug(cl_ord_id)
        return cl_ord_id


# Alias
BrokerBack = BackBroker



import traceback

from aq_lib.event.event_type import EventType
from aq_lib.model.strategy_status import StrategyStatus

from arquants.definitions import aqhandler
from arquants import logger


class Directives:
    def __init__(self):
        aqhandler.register_callback(EventType.STRATEGY_STATUS_CHANGE.value, self.process_stop)
        aqhandler.register_callback(EventType.MANUAL_ORDER_CANCEL_REQUEST_EVENT.value, self.manual_cancel)

    def process_stop(self, evt):
        try:
            if evt.status == StrategyStatus.FINISHED.value:
                self.cerebro.runstop()
            else:
                self.cerebro.runningstrats[0].process_status_update(evt)
        except:
            just_the_string = traceback.format_exc()
            logger.debug(just_the_string)
            logger.exception("hubo un error al cancelar")

    def manual_cancel(self, evt):
        self.cerebro.broker.manual_cancel(evt.order_id)


directives = Directives()



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
###############################################################################
#
# Copyright (C) 2015, 2016, 2017 Daniel Rodriguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################
from __future__ import (absolute_import, division, print_function, unicode_literals)

import collections
import itertools
import json
import random
import threading
import uuid
from copy import copy
from datetime import datetime, timedelta
import traceback
from aq_lib.event.event_type import EventType
from aq_lib.model.md_subscription_type import MDSubscriptionType
from aq_lib.utils.logger import logger
from backtrader.metabase import MetaParams
from backtrader.utils import AutoDict
from backtrader.utils.py3 import bstr, queue, with_metaclass, long
from arquants.definitions import aqhandler
from arquants.event.event_factory import eventFactory
from arquants.observers.position_calculator import PositionCalculator

bytes = bstr  # py2/3 need for ibpy


def _ts2dt(tstamp=None):
    # Transforms a RTVolume timestamp to a datetime object
    if not tstamp:
        return datetime.utcnow()

    sec, msec = divmod(long(tstamp), 1000)
    usec = msec * 1000
    return datetime.utcfromtimestamp(sec).replace(microsecond=usec)


class RTVolume(object):
    """Parses a tickString tickType 48 (RTVolume) event from the IB API into its
    constituent fields

    Supports using a "price" to simulate an RTVolume from a tickPrice event
    """
    _fields = [
        ('price', float),
        ('size', int),
        ('datetime', _ts2dt),
        ('volume', int),
        ('vwap', float),
        ('single', bool)
    ]

    def __init__(self, rtvol='', price=None, tmoffset=None):
        # Use a provided string or simulate a list of empty tokens
        tokens = iter(rtvol.split(';'))

        # Put the tokens as attributes using the corresponding func
        for name, func in self._fields:
            setattr(self, name, func(next(tokens)) if rtvol else func())

        # If price was provided use it
        if price is not None:
            self.price = price

        if tmoffset is not None:
            self.datetime += tmoffset


class MetaSingleton(MetaParams):
    """Metaclass to make a metaclassed class a singleton"""

    def __init__(cls, name, bases, dct):
        super(MetaSingleton, cls).__init__(name, bases, dct)
        cls._singleton = None

    def __call__(cls, *args, **kwargs):
        if cls._singleton is None:
            cls._singleton = (
                super(MetaSingleton, cls).__call__(*args, **kwargs))

        return cls._singleton


class AQStore(with_metaclass(MetaSingleton, object)):
    """Singleton class wrapping an ibpy ibConnection instance.

    The parameters can also be specified in the classes which use this store,
    like ``IBData`` and ``IBBroker``

    Params:

      - ``notifyall`` (default: ``False``)

        If ``False`` only ``error`` messages will be sent to the
        ``notify_store`` methods of ``Cerebro`` and ``Strategy``.

        If ``True``, each and every message received from TWS will be notified

      - ``_debug`` (default: ``False``)

        Print all messages received from TWS to standard output

      - ``timeoffset`` (default: ``True``)

        If True, the time obtained from ``reqCurrentTime`` (IB Server time)
        will be used to calculate the offset to localtime and this offset will
        be used for the price notifications (tickPrice events, for example for
        CASH markets) to modify the locally calculated timestamp.

        The time offset will propagate to other parts of the ``backtrader``
        ecosystem like the **resampling** to align resampling timestamps using
        the calculated offset.

      - ``timerefresh`` (default: ``60.0``)

        Time in seconds: how often the time offset has to be refreshed

    """

    # Set a base for the data requests (historical/realtime) to distinguish the
    # id in the error notifications from orders, where the basis (usually
    # starting at 1) is set by TWS
    REQIDBASE = 0x01000000

    BrokerCls = None  # broker class will autoregister
    DataCls = None  # data class will auto register

    datas_class = dict()

    params = (
        ('notifyall', False),
        ('_debug', False),
        ('clientId', None),
        ('timeoffset', True),  # Use offset to server for timestamps if needed
        ('timerefresh', 60.0),  # How often to refresh the timeoffset
    )

    @classmethod
    def getdata(cls, *args, **kwargs):
        """Returns ``DataCls`` with args, kwargs"""
        return cls.DataCls(*args, **kwargs)

    @classmethod
    def getdataclass(cls, _type, *args, **kwargs):
        return cls.datas_class[_type]

    @classmethod
    def getbroker(cls, *args, **kwargs):
        """Returns broker with *args, **kwargs from registered ``BrokerCls``"""
        return cls.BrokerCls(*args, **kwargs)

    def __init__(self, async_queue):
        super(AQStore, self).__init__()
        self.queue = async_queue

        self._lock_q = threading.Lock()  # sync access to _tickerId/Queues
        self._lock_accupd = threading.Lock()  # sync account updates       #TODO Definir si es necesario
        self._lock_pos = threading.Lock()  # sync account updates          #TODO Definir si es necesario
        self._lock_notif = threading.Lock()  # sync access to notif queue

        # Account list received
        self._event_managed_accounts = threading.Event()  # TODO Definir si es necesario
        self._event_accdownload = threading.Event()  # TODO Definir si es necesario

        self._env = None  # reference to cerebro for general notifications
        self.broker = None  # broker instance
        self.datas = list()  # datas that have registered over start
        self.ccount = 0  # requests to start (from cerebro or datas)       #TODO Definir si es necesario

        self._lock_tmoffset = threading.Lock()  # TODO Definir si es necesario
        self.tmoffset = timedelta()  # to control time difference with server    #TODO Definir si es necesario

        # Structures to hold datas requests
        self.qs = collections.OrderedDict()  # key: tickerId -> queues
        self.ts = collections.OrderedDict()  # key: queue -> tickerId
        self.subsbysymbol = collections.OrderedDict()  # key: symbol -> tickerIds

        self.iscash = dict()  # tickerIds from cash products (for ex: EUR.JPY)

        self.histexreq = dict()  # holds segmented historical requests
        self.histfmt = dict()  # holds datetimeformat for request
        self.histsend = dict()  # holds sessionend (data time) for request
        self.histtz = dict()  # holds sessionend (data time) for request

        self.acc_cash = AutoDict()  # current total cash per account
        self.acc_value = AutoDict()  # current total value per account
        self.acc_upds = AutoDict()  # current account valueinfos per account

        self.port_update = False  # indicate whether to signal to broker

        self.positions = collections.defaultdict(PositionCalculator)  # actual positions

        self._tickerId = itertools.count(self.REQIDBASE)  # unique tickerIds
        self.orderid = None  # next possible orderid (will be itertools.count)

        self.cdetails = collections.defaultdict(list)  # hold cdetails requests

        self.managed_accounts = list()  # received via managedAccounts

        self.notifs = queue.Queue()  # store notifications for cerebro

        # This utility key function transforms a barsize into a:
        #   (Timeframe, Compression) tuple which can be sorted
        def keyfn(x):
            n, t = x.split()
            tf, comp = self._sizes[t]
            return tf, int(n) * comp

        # This utility key function transforms a duration into a:
        #   (Timeframe, Compression) tuple which can be sorted
        def key2fn(x):
            n, d = x.split()
            tf = self._dur2tf[d]
            return tf, int(n)

        aqhandler.register_callback(EventType.MARKET_DATA_SNAPSHOT_FULL_REFRESH_EVENT.value, self.md_callback)
        aqhandler.register_callback(EventType.EXECUTION_REPORT_EVENT.value, self.exec_details)
        # self.routing_key = definitions.configuration.get_property("RABBITMQ", "OutcomeEventRoutingKey")
        if self.p.clientId is None:
            self.clientId = random.randint(1, pow(2, 16) - 1)
        else:
            self.clientId = self.p.clientId

    def start(self, data=None, broker=None):
        if data is not None:
            self._env = data._env
            self.datas.append(data)
            return self.req_mkt_data(data)
        elif broker is not None:
            self.broker = broker

    # TODO Definir que debe hacer este metodo
    def stop(self):
        # Unblock any calls set on these events
        self._event_managed_accounts.set()
        self._event_accdownload.set()

    def logmsg(self, *args):
        # for logging purposes
        if self.p._debug:
            logger.debug(*args)

    # TODO Definir que debe hacer este metodo
    def watcher(self, msg):
        # will be registered to see all messages if debug is requested
        self.logmsg(str(msg))
        if self.p.notifyall:
            self.notifs.put((msg, tuple(msg.values()), dict(msg.items())))

    # TODO Chequear si esto debe hacerse para todas las datas juntas
    def startdatas(self):
        # kickstrat datas, not returning until all of them have been done
        ts = list()
        for data in self.datas:
            t = threading.Thread(target=data.reqdata)
            t.start()
            ts.append(t)

        for t in ts:
            t.join()

    # TODO Chequear si esto debe hacerse para todas las datas juntas
    def stopdatas(self):
        # stop subs and force datas out of the loop (in LIFO order)
        qs = list(self.qs.values())
        ts = list()
        for data in self.datas:
            t = threading.Thread(target=data.canceldata)
            t.start()
            ts.append(t)

        for t in ts:
            t.join()

        for q in reversed(qs):  # datamaster the last one to get a None
            q.put(None)

    # TODO Ver quien usa este metodo y si es necesario
    def get_notifications(self):
        """Return the pending "store" notifications"""
        # The background thread could keep on adding notifications. The None
        # mark allows to identify which is the last notification to deliver
        self.notifs.put(None)  # put a mark
        notifs = list()
        while True:
            notif = self.notifs.get()
            if notif is None:  # mark is reached
                break
            notifs.append(notif)

        return notifs

    def get_ticker_queue(self, data_type, contract):
        q = queue.Queue(maxsize=1)
        with self._lock_q:
            symbol = ''.join(e for e in contract if e.isalnum())
            ticker_id = "MD-" + str(self.generate_id())
            key = data_type + "-" + symbol
            if not self.subsbysymbol.__contains__(key):
                self.subsbysymbol[key] = list()
            self.subsbysymbol[key].append(ticker_id)
            self.qs[ticker_id] = q
            self.ts[q] = ticker_id
        return ticker_id, q

    def cancel_queue(self, q, sendnone=False):
        """Cancels a Queue for data delivery"""
        # pop ts (tickers) and with the result qs (queues)
        ticker_id = self.ts.pop(q, None)
        self.qs.pop(ticker_id, None)

        self.iscash.pop(ticker_id, None)

        if sendnone:
            q.put(None)

    def valid_queue(self, q):
        """Returns (bool)  if a queue is still valid"""
        return q in self.ts  # queue -> ticker

    def req_mkt_data(self, data):
        md_request_id, q = self.get_ticker_queue(data.data_type, data.tradecontract)
        self.publish_market_data_request(data, md_request_id)
        return q

    @staticmethod
    def publish_market_data_request(data, md_request_id):
        logger.debug("Sending MD request - Symbol: %s - MDRequestID: %s" % (data.tradecontract, md_request_id))
        md_type = None
        if data.data_type is 'aggregated':
            md_type = MDSubscriptionType.AGGREGATED.value
        elif data.data_type is 'disaggregated':
            md_type = MDSubscriptionType.DISAGGREGATED.value
        elif data.data_type is 'top':
            md_type = MDSubscriptionType.TOP.value

        a = json.dumps(eventFactory.createMarketDataRequestEvent(contract=data.tradecontract,
                                                                 settlement_date=data.settlement_date,
                                                                 security_type=data.security_type,
                                                                 currency=data.currency,
                                                                 settlement=data.settlement,
                                                                 md_request_id=md_request_id,
                                                                 data_type=md_type,
                                                                 market=data.market,
                                                                 negotiation_agent_id=data.negotiation_agent_id
                                                                 ).__dict__)
        aqhandler.send(a, data.routingkey)

    def cancel_mkt_data(self, q):
        """Cancels an existing MarketData subscription

        Params:
          - q: the Queue returned by reqMktData
        """
        with self._lock_q:
            ticker_id = self.ts.get(q, None)
            if ticker_id is not None:
                logger.debug("TODO: Aca hay que enviar un evento al bus para desubscribirse")
            self.cancel_queue(q, True)

    def md_callback(self, md):

        logger.info("Llegó MD")
        s = ''.join(e for e in md.symbol if e.isalnum())
        if md.md_type is MDSubscriptionType.DISAGGREGATED.value:
            prefix = 'disaggregated-'
        elif md.md_type is MDSubscriptionType.AGGREGATED.value:
            prefix = 'aggregated-'
        elif md.md_type is MDSubscriptionType.TOP.value:
            prefix = 'top-'
        symbol = prefix + s
        if self.subsbysymbol.__contains__(symbol):
            keys = self.subsbysymbol[symbol]
            for key in keys:
                if self.qs.__contains__(key):
                    if self.qs[key].full():
                        try:
                            self.qs[key].get(block=False)
                        except queue.Empty:
                            pass
                    self.qs[key].put(md)
                    # Disparar señal para activar cerebro
                    self.queue.put('Llego algo')
                else:
                    logger.debug("The queue doesn't have the symbol %s" % key)
        else:
            logger.debug("Doesn't exists a subscription for this symbol")

    @staticmethod
    def cancel_order(order):
        try:
            aqhandler.send(json.dumps(eventFactory.createCancelOrderEventFromOrder(order).to_dict()),
                           order.data.routingkey)
        except Exception as e:
            just_the_string = traceback.format_exc()
            logger.debug(just_the_string)
            logger.exception("cancel_order: Error processing order cancel", e)

    def mass_cancel(self, routing_key=None, segment=None, account=None):
        evt = eventFactory.createOrderMassCancelRequestEvent(segment=segment, account=account,
                                                             clientOrderId=self.generate_id()).to_dict()
        aqhandler.send(json.dumps(evt), routing_key)

    # TODO esto deberia soportar multiples mercados. Va por el lado de la routing key
    def send_bulk_new_orders(self, orders):
        aqhandler.send(json.dumps(eventFactory.createBulkNewOrderEvent(orders).to_dict()), orders[0].data.routingkey)
        logger.info("send_bulk_new_orders: publishing orders in bulk")
        for order in orders:
            aqhandler.send(json.dumps(eventFactory.createOrderUpdatedEventFromOrder(order).to_dict()),
                           order.data.routingkey)

    @staticmethod
    def place_order(order):
        if getattr(order, 'm_send', True):
            if hasattr(order, 'm_originalClientOrderId'):
                aqhandler.send(json.dumps(eventFactory.createOrderCancelReplaceEvent(order).to_dict()),
                               order.data.routingkey)
            else:
                aqhandler.send(json.dumps(eventFactory.createNewOrderEventFromOrder(order).to_dict()),
                               order.data.routingkey)

    def exec_details(self, msg):
        try:
            self.broker.push_execution(msg)
            self.queue.put('exec_details: received execution report ')
        except Exception as e:
            just_the_string = traceback.format_exc()
            logger.debug(just_the_string)
            logger.exception("exec_details: exception processing execution report", e)

    # TODO Deberia transformarse en un metodo que setea la posicion inicial a partir de lo que envia ROFEX
    def update_portfolio(self, msg):
        # Lock access to the position dicts. This is called in sub-thread and
        # can kick in at any time
        with self._lock_pos:
            if not self._event_accdownload.is_set():  # 1st event seen
                position = PositionCalculator()
                self.positions[msg.contract.m_conId] = position
            else:
                position = self.positions[msg.contract.m_conId]
                if not position.fix(msg.position, msg.averageCost):
                    err = ('The current calculated position and '
                           'the position reported by the broker do not match. '
                           'Operation can continue, but the trades '
                           'calculated in the strategy may be wrong')

                    self.notifs.put((err, (), {}))

                # Flag signal to broker at the end of account download
                # self.port_update = True
                self.broker.push_portupdate()

    def getposition(self, contract, clone=False):
        # Lock access to the position dicts. This is called from main thread
        # and updates could be happening in the background
        with self._lock_pos:
            position = self.positions[contract]
            if clone:
                return copy(position)

            return position

    def get_delivery_rk(self):
        return self._env.delivery_routing_key

    @staticmethod
    def generate_id():
        cl_ord_id = uuid.uuid4().hex[:15]
        return cl_ord_id



import math
import re


class RequestGenerator:

    def __init__(self):
        self.types = {
            int: 'numberValue',
            float: 'numberValue',
            str: "stringValue",
            bool: 'boolValue'
        }

    def gen_json_update_request(self, rows, sheet_id, a1_position):
        json_rows = list()
        for row in rows:
            json_rows.append(self.gen_json_row(row))

        x, y = self.get_xy_coordinates(a1_position)

        return {
            "updateCells": {
                "rows": json_rows,
                "fields": '*',
                'start': {
                    "sheetId": sheet_id,
                    "rowIndex": y,
                    "columnIndex": x
                }
            }
        }

    @staticmethod
    def _convert_a_to_x(value):
        total = 0
        for i in range(1, len(value) + 1):
            letter = value[-1 * i]
            val = ord(letter) - ord('A')
            if i > 1:
                val = math.pow(26, (i-1)) * (val +1)
            total += val

        total = int(total)
        return total

    @staticmethod
    def _split_a1(value):
        result = re.match('[A-Z]*', value)
        if not result:
            return None
        col_value = result.group()
        row_value = value[result.end():]
        return col_value, row_value

    def get_xy_coordinates(self, value):
        col, row = self._split_a1(value)
        x = self._convert_a_to_x(col)
        y = int(row) - 1
        return x, y

    def gen_json_append_request(self, rows, sheet_id):
        json_rows = list()
        for row in rows:
            json_rows.append(self.gen_json_row(row))

        return {
            "appendCells": {
                "sheetId": sheet_id,
                "rows": json_rows,
                "fields": '*'
            }
        }

    def gen_json_row(self, values):
        json_values = list()
        for value in values:
            json_values.append(self.gen_json_value(value))
        return {'values': json_values}

    def gen_json_value(self, value):
        return {'userEnteredValue': {self.types.get(type(value)): str(value)}}


# TEST HARNESS =================================================================
# run_test = False
#
# if run_test:
#     generator = RequestGenerator()
#
#     result = generator.gen_json_append_request([
#         ['cadillac', 'eldorado', 1973, 30000],
#         ['maserati', 'GT', 80000]],
#         1256842794
#     )
#     print(json.dumps(result, sort_keys=True, indent=4))
#
#     result = generator.gen_json_update_request([
#         ['Peugeot', '308 SW'],
#         ['Peugeot', '406']],
#         1256842794,
#         'A4'
#     )
#     print(json.dumps(result, sort_keys=True, indent=4))



import logging


def init_logging(name_logger):
    """
    Configuro logging para llevar registro del AQhandler.
    """
    aq_logger = logging.getLogger(name_logger)
    aq_logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s')

    fh = logging.FileHandler('AQHandler.log')
    fh.setFormatter(formatter)
    aq_logger.addHandler(fh)

    return aq_logger


logger = init_logging('aq_handler')
